# -*- coding: utf-8 -*-
"""
    wakatime.project
    ~~~~~~~~~~~~~~~~

    Returns a project for the given file.

    :copyright: (c) 2013 Alan Hamlett.
    :license: BSD, see LICENSE for more details.
"""

import logging

from .projects.git import Git
from .projects.mercurial import Mercurial
from .projects.projectfile import ProjectFile
from .projects.projectmap import ProjectMap
from .projects.subversion import Subversion


log = logging.getLogger('WakaTime')


# List of plugin classes to find a project for the current file path.
CONFIG_PLUGINS = [
    ProjectFile,
    ProjectMap,
]
REV_CONTROL_PLUGINS = [
    Git,
    Mercurial,
    Subversion,
]


def get_project_info(configs, heartbeat, data):
    """Find the current project and branch.

    First looks for a .wakatime-project file. Second, uses the --project arg.
    Third, uses the folder name from a revision control repository. Last, uses
    the --alternate-project arg.

    Returns a project, branch tuple.
    """

    project_name, branch_name = heartbeat.project, heartbeat.branch

    if heartbeat.type != 'file':
        project_name = project_name or heartbeat.args.project or heartbeat.args.alternate_project
        return project_name, branch_name

    if project_name is None or branch_name is None:

        for plugin_cls in CONFIG_PLUGINS:

            plugin_name = plugin_cls.__name__.lower()
            plugin_configs = get_configs_for_plugin(plugin_name, configs)

            project = plugin_cls(heartbeat.entity, configs=plugin_configs)
            if project.process():
                project_name = project_name or project.name()
                branch_name = project.branch()
                break

    if project_name is None:
        project_name = data.get('project') or heartbeat.args.project

    if project_name is None or branch_name is None:

        for plugin_cls in REV_CONTROL_PLUGINS:

            plugin_name = plugin_cls.__name__.lower()
            plugin_configs = get_configs_for_plugin(plugin_name, configs)

            project = plugin_cls(heartbeat.entity, configs=plugin_configs)
            if project.process():
                project_name = project_name or project.name()
                branch_name = branch_name or project.branch()
                break

    if project_name is None:
        project_name = data.get('alternate_project') or heartbeat.args.alternate_project

    return project_name, branch_name


def get_configs_for_plugin(plugin_name, configs):
    if configs and configs.has_section(plugin_name):
        return dict(configs.items(plugin_name))
    return None



# This file is part of 'NTLM Authorization Proxy Server' http://sourceforge.net/projects/ntlmaps/
# Copyright 2001 Dmitry A. Rozmanov <dima@xenon.spb.ru>
#
# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see <http://www.gnu.org/licenses/> or <http://www.gnu.org/licenses/lgpl.txt>.
import six

from ntlm_auth.U32 import U32
from ntlm_auth.des_data import des_SPtrans, des_skb

def c2l(c):
    "char[4] to unsigned long"
    l = U32(c[0])
    l = l | (U32(c[1]) << 8)
    l = l | (U32(c[2]) << 16)
    l = l | (U32(c[3]) << 24)
    return l


def l2c(l):
    "unsigned long to char[4]"
    c = []
    c.append(int(l & U32(0xFF)))
    c.append(int((l >> 8) & U32(0xFF)))
    c.append(int((l >> 16) & U32(0xFF)))
    c.append(int((l >> 24) & U32(0xFF)))
    return c


def D_ENCRYPT(tup, u, t, s):
    L, R, S = tup
    # print 'LRS1', L, R, S, u, t, '-->',
    u = (R ^ s[S])
    t = R ^ s[S + 1]
    t = ((t >> 4) + (t << 28))
    L = L ^ (des_SPtrans[1][int((t) & U32(0x3f))] |
             des_SPtrans[3][int((t >> 8) & U32(0x3f))] |
             des_SPtrans[5][int((t >> 16) & U32(0x3f))] |
             des_SPtrans[7][int((t >> 24) & U32(0x3f))] |
             des_SPtrans[0][int((u) & U32(0x3f))] |
             des_SPtrans[2][int((u >> 8) & U32(0x3f))] |
             des_SPtrans[4][int((u >> 16) & U32(0x3f))] |
             des_SPtrans[6][int((u >> 24) & U32(0x3f))])
    # print 'LRS:', L, R, S, u, t
    return (L, R, S), u, t, s


def PERM_OP(tup, n, m):
    "tup - (a, b, t)"
    a, b, t = tup
    t = ((a >> n) ^ b) & m
    b = b ^ t
    a = a ^ (t << n)
    return (a, b, t)


def HPERM_OP(tup, n, m):
    "tup - (a, t)"
    a, t = tup
    t = ((a << (16 - n)) ^ a) & m
    a = a ^ t ^ (t >> (16 - n))
    return a, t


shifts2 = [0, 0, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 0]


class DES:
    KeySched = None  # des_key_schedule

    def __init__(self, key_str):
        self.KeySched = des_set_key(key_str)

    def decrypt(self, str):
        # block - UChar[]
        block = []

        for i in six.iterbytes(str):
            block.append(i)

        # print block
        block = des_ecb_encrypt(block, self.KeySched, 0)

        res = b''
        for i in block:
            res = res + six.int2byte(i)

        return res

    def encrypt(self, plaintext):
        # block - UChar[]

        block = []
        for i in plaintext:
            block.append(i)

        block = des_ecb_encrypt(block, self.KeySched, 1)

        res = b''

        for i in block:
            res += six.int2byte(i)

        return res


def des_encript(input, ks, encrypt):
    # input - U32[]
    # output - U32[]
    # ks - des_key_shedule - U32[2][16]
    # encrypt - int
    # l, r, t, u - U32
    # i - int
    # s - U32[]

    l = input[0]
    r = input[1]
    t = U32(0)
    u = U32(0)

    r, l, t = PERM_OP((r, l, t), 4, U32(0x0f0f0f0f))
    l, r, t = PERM_OP((l, r, t), 16, U32(0x0000ffff))
    r, l, t = PERM_OP((r, l, t), 2, U32(0x33333333))
    l, r, t = PERM_OP((l, r, t), 8, U32(0x00ff00ff))
    r, l, t = PERM_OP((r, l, t), 1, U32(0x55555555))

    t = (r << 1) | (r >> 31)
    r = (l << 1) | (l >> 31)
    l = t

    s = ks  # ???????????????
    # print l, r
    if encrypt:
        for i in range(0, 32, 4):
            rtup, u, t, s = D_ENCRYPT((l, r, i + 0), u, t, s)
            l = rtup[0]
            r = rtup[1]
            rtup, u, t, s = D_ENCRYPT((r, l, i + 2), u, t, s)
            r = rtup[0]
            l = rtup[1]
    else:
        for i in range(30, 0, -4):
            rtup, u, t, s = D_ENCRYPT((l, r, i - 0), u, t, s)
            l = rtup[0]
            r = rtup[1]
            rtup, u, t, s = D_ENCRYPT((r, l, i - 2), u, t, s)
            r = rtup[0]
            l = rtup[1]
        # print l, r
    l = (l >> 1) | (l << 31)
    r = (r >> 1) | (r << 31)

    r, l, t = PERM_OP((r, l, t), 1, U32(0x55555555))
    l, r, t = PERM_OP((l, r, t), 8, U32(0x00ff00ff))
    r, l, t = PERM_OP((r, l, t), 2, U32(0x33333333))
    l, r, t = PERM_OP((l, r, t), 16, U32(0x0000ffff))
    r, l, t = PERM_OP((r, l, t), 4, U32(0x0f0f0f0f))

    output = [l]
    output.append(r)
    l, r, t, u = U32(0), U32(0), U32(0), U32(0)
    return output


def des_ecb_encrypt(input, ks, encrypt):
    # input - des_cblock - UChar[8]
    # output - des_cblock - UChar[8]
    # ks - des_key_shedule - U32[2][16]
    # encrypt - int

    # print input
    l0 = c2l(input[0:4])
    l1 = c2l(input[4:8])
    ll = [l0]
    ll.append(l1)
    # print ll
    ll = des_encript(ll, ks, encrypt)
    # print ll
    l0 = ll[0]
    l1 = ll[1]
    output = l2c(l0)
    output = output + l2c(l1)
    # print output
    l0, l1, ll[0], ll[1] = U32(0), U32(0), U32(0), U32(0)
    return output


def des_set_key(key):
    # key - des_cblock - UChar[8]
    # schedule - des_key_schedule

    # register unsigned long c,d,t,s;
    # register unsigned char *in;
    # register unsigned long *k;
    # register int i;

    # k = schedule
    # in = key

    k = []
    c = c2l(key[0:4])
    d = c2l(key[4:8])
    t = U32(0)

    d, c, t = PERM_OP((d, c, t), 4, U32(0x0f0f0f0f))
    c, t = HPERM_OP((c, t), -2, U32(0xcccc0000))
    d, t = HPERM_OP((d, t), -2, U32(0xcccc0000))
    d, c, t = PERM_OP((d, c, t), 1, U32(0x55555555))
    c, d, t = PERM_OP((c, d, t), 8, U32(0x00ff00ff))
    d, c, t = PERM_OP((d, c, t), 1, U32(0x55555555))

    d = (((d & U32(0x000000ff)) << 16) | (d & U32(0x0000ff00)) | ((d & U32(0x00ff0000)) >> 16) | (
        (c & U32(0xf0000000)) >> 4))
    c = c & U32(0x0fffffff)

    for i in range(16):
        if (shifts2[i]):
            c = ((c >> 2) | (c << 26))
            d = ((d >> 2) | (d << 26))
        else:
            c = ((c >> 1) | (c << 27))
            d = ((d >> 1) | (d << 27))
        c = c & U32(0x0fffffff)
        d = d & U32(0x0fffffff)

        s = des_skb[0][int((c) & U32(0x3f))] | \
            des_skb[1][int(((c >> 6) & U32(0x03)) | ((c >> 7) & U32(0x3c)))] | \
            des_skb[2][int(((c >> 13) & U32(0x0f)) | ((c >> 14) & U32(0x30)))] | \
            des_skb[3][int(((c >> 20) & U32(0x01)) | ((c >> 21) & U32(0x06)) | ((c >> 22) & U32(0x38)))]

        t = des_skb[4][int((d) & U32(0x3f))] | \
            des_skb[5][int(((d >> 7) & U32(0x03)) | ((d >> 8) & U32(0x3c)))] | \
            des_skb[6][int((d >> 15) & U32(0x3f))] | \
            des_skb[7][int(((d >> 21) & U32(0x0f)) | ((d >> 22) & U32(0x30)))]
        # print s, t

        k.append(((t << 16) | (s & U32(0x0000ffff))) & U32(0xffffffff))
        s = ((s >> 16) | (t & U32(0xffff0000)))
        s = (s << 4) | (s >> 28)
        k.append(s & U32(0xffffffff))

    schedule = k

    return schedule



# -*- coding: utf-8 -*-
"""
    wakatime.languages.c_cpp
    ~~~~~~~~~~~~~~~~~~~~~~~~

    Parse dependencies from C++ code.

    :copyright: (c) 2014 Alan Hamlett.
    :license: BSD, see LICENSE for more details.
"""

from . import TokenParser


class CParser(TokenParser):
    exclude = [
        r'^stdio\.h$',
        r'^stdlib\.h$',
        r'^string\.h$',
        r'^time\.h$',
    ]
    state = None

    def parse(self):
        for index, token, content in self.tokens:
            self._process_token(token, content)
        return self.dependencies

    def _process_token(self, token, content):
        if self.partial(token) == 'Preproc' or self.partial(token) == 'PreprocFile':
            self._process_preproc(token, content)
        else:
            self._process_other(token, content)

    def _process_preproc(self, token, content):
        if self.state == 'include':
            if content != '\n' and content != '#':
                content = content.strip().strip('"').strip('<').strip('>').strip()
                self.append(content, truncate=True, separator='/')
            self.state = None
        elif content.strip().startswith('include'):
            self.state = 'include'
        else:
            self.state = None

    def _process_other(self, token, content):
        pass


class CppParser(CParser):
    pass



# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see <http://www.gnu.org/licenses/> or <http://www.gnu.org/licenses/lgpl.txt>.

import base64
import socket
import struct
from ntlm_auth.constants import NegotiateFlags
from ntlm_auth.messages import NegotiateMessage, ChallengeMessage, AuthenticateMessage
from ntlm_auth.session_security import SessionSecurity


"""
utility functions for Microsoft NTLM authentication

References:
[MS-NLMP]: NT LAN Manager (NTLM) Authentication Protocol Specification
http://download.microsoft.com/download/a/e/6/ae6e4142-aa58-45c6-8dcf-a657e5900cd3/%5BMS-NLMP%5D.pdf

[MS-NTHT]: NTLM Over HTTP Protocol Specification
http://download.microsoft.com/download/a/e/6/ae6e4142-aa58-45c6-8dcf-a657e5900cd3/%5BMS-NTHT%5D.pdf

Cntlm Authentication Proxy
http://cntlm.awk.cz/

NTLM Authorization Proxy Server
http://sourceforge.net/projects/ntlmaps/

Optimized Attack for NTLM2 Session Response
http://www.blackhat.com/presentations/bh-asia-04/bh-jp-04-pdfs/bh-jp-04-seki.pdf
"""

class Ntlm(object):
    """
    Initialises the NTLM context to use when sending and receiving messages to and from the server. You should be
    using this object as it supports NTLMv2 authenticate and it easier to use than before. It also brings in the
    ability to use signing and sealing with session_security and generate a MIC structure.

    :param ntlm_compatibility: The Lan Manager Compatibility Level to use withe the auth message - Default 3
                                This is set by an Administrator in the registry key
                                'HKLM\SYSTEM\CurrentControlSet\Control\Lsa\LmCompatibilityLevel'
                                The values correspond to the following;
                                    0 : LM and NTLMv1
                                    1 : LM, NTLMv1 and NTLMv1 with Extended Session Security
                                    2 : NTLMv1 and NTLMv1 with Extended Session Security
                                    3-5 : NTLMv2 Only
                                Note: Values 3 to 5 are no different as the client supports the same types

    Attributes:
        negotiate_flags: A NEGOTIATE structure that contains a set of bit flags. These flags are the options the client supports and are sent in the negotiate_message
        ntlm_compatibility: The Lan Manager Compatibility Level, same as the input if supplied
        negotiate_message: A NegotiateMessage object that is sent to the server
        challenge_message: A ChallengeMessage object that has been created from the server response
        authenticate_message: An AuthenticateMessage object that is sent to the server based on the ChallengeMessage
        session_security: A SessionSecurity structure that can be used to sign and seal messages sent after the authentication challenge
    """
    def __init__(self, ntlm_compatibility=3):
        self.ntlm_compatibility = ntlm_compatibility

        # Setting up our flags so the challenge message returns the target info block if supported
        self.negotiate_flags = NegotiateFlags.NTLMSSP_NEGOTIATE_TARGET_INFO | \
                               NegotiateFlags.NTLMSSP_NEGOTIATE_128 | \
                               NegotiateFlags.NTLMSSP_NEGOTIATE_56 | \
                               NegotiateFlags.NTLMSSP_NEGOTIATE_UNICODE | \
                               NegotiateFlags.NTLMSSP_NEGOTIATE_VERSION | \
                               NegotiateFlags.NTLMSSP_NEGOTIATE_KEY_EXCH | \
                               NegotiateFlags.NTLMSSP_NEGOTIATE_ALWAYS_SIGN | \
                               NegotiateFlags.NTLMSSP_NEGOTIATE_SIGN | \
                               NegotiateFlags.NTLMSSP_NEGOTIATE_SEAL

        # Setting the message types based on the ntlm_compatibility level
        self._set_ntlm_compatibility_flags(self.ntlm_compatibility)

        self.negotiate_message = None
        self.challenge_message = None
        self.authenticate_message = None
        self.session_security = None


    def create_negotiate_message(self, domain_name=None, workstation=None):
        """
        Create an NTLM NEGOTIATE_MESSAGE

        :param domain_name: The domain name of the user account we are authenticating with, default is None
        :param worksation: The workstation we are using to authenticate with, default is None
        :return: A base64 encoded string of the NEGOTIATE_MESSAGE
        """
        self.negotiate_message = NegotiateMessage(self.negotiate_flags, domain_name, workstation)

        return base64.b64encode(self.negotiate_message.get_data())

    def parse_challenge_message(self, msg2):
        """
        Parse the NTLM CHALLENGE_MESSAGE from the server and add it to the Ntlm context fields

        :param msg2: A base64 encoded string of the CHALLENGE_MESSAGE
        """
        msg2 = base64.b64decode(msg2)
        self.challenge_message = ChallengeMessage(msg2)

    def create_authenticate_message(self, user_name, password, domain_name=None, workstation=None, server_certificate_hash=None):
        """
        Create an NTLM AUTHENTICATE_MESSAGE based on the Ntlm context and the previous messages sent and received

        :param user_name: The user name of the user we are trying to authenticate with
        :param password: The password of the user we are trying to authenticate with
        :param domain_name: The domain name of the user account we are authenticated with, default is None
        :param workstation: The workstation we are using to authenticate with, default is None
        :param server_certificate_hash: The SHA256 hash string of the server certificate (DER encoded) NTLM is authenticating to. Used for Channel
                                        Binding Tokens. If nothing is supplied then the CBT hash will not be sent. See messages.py AuthenticateMessage
                                        for more details
        :return: A base64 encoded string of the AUTHENTICATE_MESSAGE
        """
        self.authenticate_message = AuthenticateMessage(user_name, password, domain_name, workstation,
                                                        self.challenge_message, self.ntlm_compatibility,
                                                        server_certificate_hash)
        self.authenticate_message.add_mic(self.negotiate_message, self.challenge_message)

        # Setups up the session_security context used to sign and seal messages if wanted
        if self.negotiate_flags & NegotiateFlags.NTLMSSP_NEGOTIATE_SEAL or self.negotiate_flags & NegotiateFlags.NTLMSSP_NEGOTIATE_SIGN:
            self.session_security = SessionSecurity(struct.unpack("<I", self.authenticate_message.negotiate_flags)[0],
                                                    self.authenticate_message.exported_session_key)

        return base64.b64encode(self.authenticate_message.get_data())

    def _set_ntlm_compatibility_flags(self, ntlm_compatibility):
        if (ntlm_compatibility >= 0) and (ntlm_compatibility <= 5):
            if ntlm_compatibility == 0:
                self.negotiate_flags |= NegotiateFlags.NTLMSSP_NEGOTIATE_NTLM | \
                                        NegotiateFlags.NTLMSSP_NEGOTIATE_LM_KEY
            elif ntlm_compatibility == 1:
                self.negotiate_flags |= NegotiateFlags.NTLMSSP_NEGOTIATE_NTLM | \
                                        NegotiateFlags.NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY
            else:
                self.negotiate_flags |= NegotiateFlags.NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY
        else:
            raise Exception("Unknown ntlm_compatibility level - expecting value between 0 and 5")



# -*- coding: utf-8 -*-
"""
    wakatime.languages.data
    ~~~~~~~~~~~~~~~~~~~~~~~

    Parse dependencies from data files.

    :copyright: (c) 2014 Alan Hamlett.
    :license: BSD, see LICENSE for more details.
"""

import os

from . import TokenParser
from ..compat import u


FILES = {
    'bower.json': {'exact': True, 'dependency': 'bower'},
    'component.json': {'exact': True, 'dependency': 'bower'},
    'package.json': {'exact': True, 'dependency': 'npm'},
}


class JsonParser(TokenParser):
    state = None
    level = 0

    def parse(self):
        self._process_file_name(os.path.basename(self.source_file))
        for index, token, content in self.tokens:
            self._process_token(token, content)
        return self.dependencies

    def _process_file_name(self, file_name):
        for key, value in FILES.items():
            found = (key == file_name) if value.get('exact') else (key.lower() in file_name.lower())
            if found:
                self.append(value['dependency'])

    def _process_token(self, token, content):
        if u(token) == 'Token.Name.Tag':
            self._process_tag(token, content)
        elif u(token) == 'Token.Literal.String.Single' or u(token) == 'Token.Literal.String.Double':
            self._process_literal_string(token, content)
        elif u(token) == 'Token.Punctuation':
            self._process_punctuation(token, content)

    def _process_tag(self, token, content):
        if content.strip('"').strip("'") == 'dependencies' or content.strip('"').strip("'") == 'devDependencies':
            self.state = 'dependencies'
        elif self.state == 'dependencies' and self.level == 2:
            self.append(content.strip('"').strip("'"))

    def _process_literal_string(self, token, content):
        pass

    def _process_punctuation(self, token, content):
        if content == '{':
            self.level += 1
        elif content == '}':
            self.level -= 1
            if self.state is not None and self.level <= 1:
                self.state = None



# -*- coding: utf-8 -*-
"""
    wakatime.cli
    ~~~~~~~~~~~~

    Command-line entry point.

    :copyright: (c) 2013 Alan Hamlett.
    :license: BSD, see LICENSE for more details.
"""

import os
import sys


# get path to local wakatime package
package_folder = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

# add local wakatime package to sys.path
sys.path.insert(0, package_folder)

# import local wakatime package
try:
    import wakatime
except (TypeError, ImportError):
    # on Windows, non-ASCII characters in import path can be fixed using
    # the script path from sys.argv[0].
    # More info at https://github.com/wakatime/wakatime/issues/32
    package_folder = os.path.dirname(os.path.dirname(os.path.abspath(sys.argv[0])))
    sys.path.insert(0, package_folder)
    import wakatime


if __name__ == '__main__':
    sys.exit(wakatime.execute(sys.argv[1:]))



# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see <http://www.gnu.org/licenses/> or <http://www.gnu.org/licenses/lgpl.txt>.

import base64
import calendar
import hashlib
import hmac
import os
import struct
import time
import ntlm_auth.compute_hash as comphash
import ntlm_auth.compute_keys as compkeys
from ntlm_auth import des
from ntlm_auth.constants import NegotiateFlags, AvFlags
from ntlm_auth.gss_channel_bindings import GssChannelBindingsStruct
from ntlm_auth.target_info import TargetInfo

class ComputeResponse():
    """
        Constructor for the response computations. This class will compute the various
        nt and lm challenge responses.

        :param user_name: The user name of the user we are trying to authenticate with
        :param password: The password of the user we are trying to authenticate with
        :param domain_name: The domain name of the user account we are authenticated with, default is None
        :param challenge_message: A ChallengeMessage object that was received from the server after the negotiate_message
        :param ntlm_compatibility: The Lan Manager Compatibility Level, used to determine what NTLM auth version to use, see Ntlm in ntlm.py for more details
    """
    def __init__(self, user_name, password, domain_name, challenge_message, ntlm_compatibility):
        self._user_name = user_name
        self._password = password
        self._domain_name = domain_name
        self._challenge_message = challenge_message
        self._negotiate_flags = challenge_message.negotiate_flags
        self._server_challenge = challenge_message.server_challenge
        self._server_target_info = challenge_message.target_info
        self._ntlm_compatibility = ntlm_compatibility
        self._client_challenge = os.urandom(8)

    def get_lm_challenge_response(self):
        """
        [MS-NLMP] v28.0 2016-07-14

        3.3.1 - NTLM v1 Authentication
        3.3.2 - NTLM v2 Authentication

        This method returns the LmChallengeResponse key based on the ntlm_compatibility chosen
        and the target_info supplied by the CHALLENGE_MESSAGE. It is quite different from what
        is set in the document as it combines the NTLMv1, NTLM2 and NTLMv2 methods into one
        and calls separate methods based on the ntlm_compatibility flag chosen.

        :return: response (LmChallengeResponse) - The LM response to the server challenge. Computed by the client
        """
        if self._negotiate_flags & NegotiateFlags.NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY and self._ntlm_compatibility < 3:
            response = ComputeResponse._get_LMv1_with_session_security_response(self._client_challenge)

        elif 0 <= self._ntlm_compatibility <= 1:
            response = ComputeResponse._get_LMv1_response(self._password, self._server_challenge)
        elif self._ntlm_compatibility == 2:
            # Based on the compatibility level we don't want to use LM responses, ignore the session_base_key as it is returned in nt
            response, ignore_key = ComputeResponse._get_NTLMv1_response(self._password, self._server_challenge)
        else:
            """
            [MS-NLMP] v28.0 page 45 - 2016-07-14

            3.1.5.12 Client Received a CHALLENGE_MESSAGE from the Server
            If NTLMv2 authentication is used and the CHALLENGE_MESSAGE TargetInfo field has an MsvAvTimestamp present,
            the client SHOULD NOT send the LmChallengeResponse and SHOULD send Z(24) instead.
            """

            response = ComputeResponse._get_LMv2_response(self._user_name, self._password, self._domain_name,
                                                          self._server_challenge,
                                                          self._client_challenge)
            if self._server_target_info is not None:
                timestamp = self._server_target_info[TargetInfo.MSV_AV_TIMESTAMP]
                if timestamp is not None:
                    response = b'\0' * 24

        return response

    def get_nt_challenge_response(self, lm_challenge_response, server_certificate_hash):
        """
        [MS-NLMP] v28.0 2016-07-14

        3.3.1 - NTLM v1 Authentication
        3.3.2 - NTLM v2 Authentication

        This method returns the NtChallengeResponse key based on the ntlm_compatibility chosen
        and the target_info supplied by the CHALLENGE_MESSAGE. It is quite different from what
        is set in the document as it combines the NTLMv1, NTLM2 and NTLMv2 methods into one
        and calls separate methods based on the ntlm_compatibility value chosen.

        :param lm_challenge_response: The LmChallengeResponse calculated beforeand, used to get the key_exchange_key value
        :param server_certificate_hash: The SHA256 hash of the server certificate (DER encoded) NTLM is authenticated to.
                                        Used in Channel Binding Tokens if present, default value is None. See
                                        AuthenticateMessage in messages.py for more details
        :return response: (NtChallengeResponse) - The NT response to the server challenge. Computed by the client
        :return session_base_key: (SessionBaseKey) - A session key calculated from the user password challenge
        :return target_info: (AV_PAIR) - The AV_PAIR structure used in the nt_challenge calculations
        """
        if self._negotiate_flags & NegotiateFlags.NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY and self._ntlm_compatibility < 3:
            # The compatibility level is less than 3 which means it doesn't support NTLMv2 but we want extended security so use NTLM2 which is different from NTLMv2
            # [MS-NLMP] - 3.3.1 NTLMv1 Authentication
            response, session_base_key = ComputeResponse._get_NTLM2_response(self._password, self._server_challenge, self._client_challenge)
            key_exchange_key = compkeys._get_exchange_key_ntlm_v1(self._negotiate_flags, session_base_key,
                                                                  self._server_challenge, lm_challenge_response,
                                                                  comphash._lmowfv1(self._password))
            target_info = None

        elif 0 <= self._ntlm_compatibility < 3:
            response, session_base_key = ComputeResponse._get_NTLMv1_response(self._password, self._server_challenge)
            key_exchange_key = compkeys._get_exchange_key_ntlm_v1(self._negotiate_flags, session_base_key,
                                                                  self._server_challenge, lm_challenge_response,
                                                                  comphash._lmowfv1(self._password))
            target_info = None
        else:
            if self._server_target_info is None:
                target_info = TargetInfo()
            else:
                target_info = self._server_target_info

            if target_info[TargetInfo.MSV_AV_TIMESTAMP] is None:
                timestamp = get_windows_timestamp()
            else:
                timestamp = target_info[TargetInfo.MSV_AV_TIMESTAMP][1]

                # [MS-NLMP] If the CHALLENGE_MESSAGE TargetInfo field has an MsvAvTimestamp present, the client SHOULD provide a MIC
                target_info[TargetInfo.MSV_AV_FLAGS] = struct.pack("<L", AvFlags.MIC_PROVIDED)

            if server_certificate_hash != None:
                channel_bindings_hash = ComputeResponse._get_channel_bindings_value(server_certificate_hash)
                target_info[TargetInfo.MSV_AV_CHANNEL_BINDINGS] = channel_bindings_hash

            response, session_base_key = ComputeResponse._get_NTLMv2_response(self._user_name, self._password, self._domain_name,
                                                 self._server_challenge, self._client_challenge, timestamp, target_info)

            key_exchange_key = compkeys._get_exchange_key_ntlm_v2(session_base_key)

        return response, key_exchange_key, target_info


    @staticmethod
    def _get_LMv1_response(password, server_challenge):
        """
        [MS-NLMP] v28.0 2016-07-14

        2.2.2.3 LM_RESPONSE
        The LM_RESPONSE structure defines the NTLM v1 authentication LmChallengeResponse
        in the AUTHENTICATE_MESSAGE. This response is used only when NTLM v1
        authentication is configured.

        :param password: The password of the user we are trying to authenticate with
        :param server_challenge: A random 8-byte response generated by the server in the CHALLENGE_MESSAGE
        :return response: LmChallengeResponse to the server challenge
        """
        lm_hash = comphash._lmowfv1(password)
        response = ComputeResponse._calc_resp(lm_hash, server_challenge)

        return response

    @staticmethod
    def _get_LMv1_with_session_security_response(client_challenge):
        """
        [MS-NLMP] v28.0 2016-07-14

        2.2.2.3 LM_RESPONSE
        The LM_RESPONSE structure defines the NTLM v1 authentication LmChallengeResponse
        in the AUTHENTICATE_MESSAGE. This response is used only when NTLM v1
        authentication is configured and NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY is flages.

        :param client_challenge: A random 8-byte response generated by the client for the AUTHENTICATE_MESSAGE
        :return response: LmChallengeResponse to the server challenge
        """

        response = client_challenge + b'\0' * 16

        return response

    @staticmethod
    def _get_LMv2_response(user_name, password, domain_name, server_challenge, client_challenge):
        """
        [MS-NLMP] v28.0 2016-07-14

        2.2.2.4 LMv2_RESPONSE
        The LMv2_RESPONSE structure defines the NTLM v2 authentication LmChallengeResponse
        in the AUTHENTICATE_MESSAGE. This response is used only when NTLM v2
        authentication is configured.

        :param user_name: The user name of the user we are trying to authenticate with
        :param password: The password of the user we are trying to authenticate with
        :param domain_name: The domain name of the user account we are authenticated with
        :param server_challenge: A random 8-byte response generated by the server in the CHALLENGE_MESSAGE
        :param client_challenge: A random 8-byte response generated by the client for the AUTHENTICATE_MESSAGE
        :return response: LmChallengeResponse to the server challenge
        """
        nt_hash = comphash._ntowfv2(user_name, password, domain_name)
        lm_hash = hmac.new(nt_hash, (server_challenge + client_challenge)).digest()
        response = lm_hash + client_challenge

        return response

    @staticmethod
    def _get_NTLMv1_response(password, server_challenge):
        """
        [MS-NLMP] v28.0 2016-07-14

        2.2.2.6 NTLM v1 Response: NTLM_RESPONSE
        The NTLM_RESPONSE strucutre defines the NTLM v1 authentication NtChallengeResponse
        in the AUTHENTICATE_MESSAGE. This response is only used when NTLM v1 authentication
        is configured.

        :param password: The password of the user we are trying to authenticate with
        :param server_challenge: A random 8-byte response generated by the server in the CHALLENGE_MESSAGE
        :return response: NtChallengeResponse to the server_challenge
        :return session_base_key: A session key calculated from the user password challenge
        """
        ntlm_hash = comphash._ntowfv1(password)
        response = ComputeResponse._calc_resp(ntlm_hash, server_challenge)

        session_base_key = hashlib.new('md4', ntlm_hash).digest()

        return response, session_base_key

    @staticmethod
    def _get_NTLM2_response(password, server_challenge, client_challenge):
        """
        [MS-NLMP] v28.0 2016-07-14

        This name is really misleading as it isn't NTLM v2 authentication rather
        This authentication is only used when the ntlm_compatibility level is set
        to a value < 3 (No NTLMv2 auth) but the NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY
        flag is set in the negotiate flags section. The documentation for computing this
        value is on page 56 under section 3.3.1 NTLM v1 Authentication

        :param password: The password of the user we are trying to authenticate with
        :param server_challenge: A random 8-byte response generated by the server in the CHALLENGE_MESSAGE
        :param client_challenge: A random 8-byte response generated by the client for the AUTHENTICATE_MESSAGE
        :return response: NtChallengeResponse to the server_challenge
        :return session_base_key: A session key calculated from the user password challenge
        """
        ntlm_hash = comphash._ntowfv1(password)
        nt_session_hash = hashlib.md5(server_challenge + client_challenge).digest()[:8]
        response = ComputeResponse._calc_resp(ntlm_hash, nt_session_hash[0:8])

        session_base_key = hashlib.new('md4', ntlm_hash).digest()

        return response, session_base_key

    @staticmethod
    def _get_NTLMv2_response(user_name, password, domain_name, server_challenge, client_challenge, timestamp, target_info):
        """
        [MS-NLMP] v28.0 2016-07-14

        2.2.2.8 NTLM V2 Response: NTLMv2_RESPONSE
        The NTLMv2_RESPONSE strucutre defines the NTLMv2 authentication NtChallengeResponse
        in the AUTHENTICATE_MESSAGE. This response is used only when NTLMv2 authentication
        is configured.

        The guide on how this is computed is in 3.3.2 NTLM v2 Authentication.

        :param user_name: The user name of the user we are trying to authenticate with
        :param password: The password of the user we are trying to authenticate with
        :param domain_name: The domain name of the user account we are authenticated with
        :param server_challenge: A random 8-byte response generated by the server in the CHALLENGE_MESSAGE
        :param client_challenge: A random 8-byte response generated by the client for the AUTHENTICATE_MESSAGE
        :param timestamp: An 8-byte timestamp in windows format, 100 nanoseconds since 1601-01-01
        :param target_info: The target_info structure from the CHALLENGE_MESSAGE with the CBT attached if required
        :return response: NtChallengeResponse to the server_challenge
        :return session_base_key: A session key calculated from the user password challenge
        """

        nt_hash = comphash._ntowfv2(user_name, password, domain_name)
        temp = ComputeResponse._get_NTLMv2_temp(timestamp, client_challenge, target_info)
        nt_proof_str = hmac.new(nt_hash, (server_challenge + temp)).digest()
        response = nt_proof_str + temp

        session_base_key = hmac.new(nt_hash, nt_proof_str).digest()

        return response, session_base_key

    @staticmethod
    def _get_NTLMv2_temp(timestamp, client_challenge, target_info):
        """
        [MS-NLMP] v28.0 2016-07-14

        2.2.2.7 NTLMv2_CLIENT_CHALLENGE - variable length
        The NTLMv2_CLIENT_CHALLENGE structure defines the client challenge in
        the AUTHENTICATE_MESSAGE. This structure is used only when NTLM v2
        authentication is configured and is transported in the NTLMv2_RESPONSE
        structure.

        The method to create this structure is defined in 3.3.2 NTLMv2 Authentication.
        In this method this variable is known as the temp value. The target_info variable
        corresponds to the ServerName variable used in that documentation. This is in
        reality a lot more than just the ServerName and contains the AV_PAIRS structure
        we need to transport with the message like Channel Binding tokens and others.
        By default this will be the target_info returned from the CHALLENGE_MESSAGE plus
        MSV_AV_CHANNEL_BINDINGS if specified otherwise it is a new target_info set with
        MSV_AV_TIMESTAMP to the current time.

        :param timestamp: An 8-byte timestamp in windows format, 100 nanoseconds since 1601-01-01
        :param client_challenge: A random 8-byte response generated by the client for the AUTHENTICATE_MESSAGE
        :param target_info: The target_info structure from the CHALLENGE_MESSAGE with the CBT attached if required
        :return temp: The CLIENT_CHALLENGE structure that will be added to the NtChallengeResponse structure
        """
        resp_type = b'\1'
        hi_resp_type = b'\1'
        reserved1 = b'\0' * 2
        reserved2 = b'\0' * 4
        reserved3 = b'\0' * 4
        reserved4 = b'\0' * 4  # This byte is not in the structure defined in 2.2.2.7 but is in the computation guide, works with it present

        temp = resp_type + hi_resp_type + reserved1 + \
               reserved2 + \
               timestamp + \
               client_challenge + \
               reserved3 + \
               target_info.get_data() + reserved4

        return temp

    @staticmethod
    def _calc_resp(password_hash, server_challenge):
        """
        Generate the LM response given a 16-byte password hash and the challenge
        from the CHALLENGE_MESSAGE

        :param password_hash: A 16-byte password hash
        :param server_challenge: A random 8-byte response generated by the server in the CHALLENGE_MESSAGE
        :return res: A 24-byte buffer to contain the LM response upon return
        """

        # padding with zeros to make the hash 21 bytes long
        password_hash += b'\0' * (21 - len(password_hash))

        res = b''
        dobj = des.DES(password_hash[0:7])
        res = res + dobj.encrypt(server_challenge[0:8])

        dobj = des.DES(password_hash[7:14])
        res = res + dobj.encrypt(server_challenge[0:8])

        dobj = des.DES(password_hash[14:21])
        res = res + dobj.encrypt(server_challenge[0:8])
        return res

    @staticmethod
    def _get_channel_bindings_value(server_certificate_hash):
        """
        https://msdn.microsoft.com/en-us/library/windows/desktop/dd919963%28v=vs.85%29.aspx?f=255&MSPPError=-2147217396
        https://blogs.msdn.microsoft.com/openspecification/2013/03/26/ntlm-and-channel-binding-hash-aka-extended-protection-for-authentication/

        Get's the MD5 hash of the gss_channel_bindings_struct to add to the AV_PAIR MSV_AV_CHANNEL_BINDINGS.
        This method takes in the SHA256 hash (Hash of the DER encoded certificate of the server we are connecting to)
        and add's it to the gss_channel_bindings_struct. It then gets the MD5 hash and converts this to a
        byte array in preparation of adding it to the AV_PAIR structure.

        :param server_certificate_hash: The SHA256 hash of the server certificate (DER encoded) NTLM is authenticated to
        :return channel_bindings: An MD5 hash of the gss_channel_bindings_struct to add to the AV_PAIR MsvChannelBindings
        """
        # Channel Binding Tokens support, used for NTLMv2
        # Decode the SHA256 certificate hash
        certificate_digest = base64.b16decode(server_certificate_hash)

        # Initialise the GssChannelBindingsStruct and add the certificate_digest to the application_data field
        gss_channel_bindings = GssChannelBindingsStruct()
        gss_channel_bindings[gss_channel_bindings.APPLICATION_DATA] = 'tls-server-end-point:'.encode() + certificate_digest

        # Get the gss_channel_bindings_struct and create an MD5 hash
        channel_bindings_struct_data = gss_channel_bindings.get_data()
        channel_bindings_hash = hashlib.md5(channel_bindings_struct_data).hexdigest()

        try:
            cbt_value = bytearray.fromhex(channel_bindings_hash)
        except TypeError:
            # Work-around for Python 2.6 bug
            cbt_value = bytearray.fromhex(unicode(channel_bindings_hash))

        channel_bindings = bytes(cbt_value)
        return channel_bindings


def get_windows_timestamp():
    # Get Windows Date time, 100 nanoseconds since 1601-01-01 in a 64 bit structure
    timestamp = struct.pack('<q', (116444736000000000 + calendar.timegm(time.gmtime()) * 10000000))

    return timestamp


import hashlib
import sys
import warnings

from ntlm_auth import ntlm
from requests.auth import AuthBase
from requests.packages.urllib3.response import HTTPResponse

class HttpNtlmAuth(AuthBase):
    """
    HTTP NTLM Authentication Handler for Requests.

    Supports pass-the-hash.
    """

    def __init__(self, username, password, session=None):
        """Create an authentication handler for NTLM over HTTP.

        :param str username: Username in 'domain\\username' format
        :param str password: Password
        :param str session: Unused. Kept for backwards-compatibility.
        """
        if ntlm is None:
            raise Exception("NTLM libraries unavailable")

        # parse the username
        try:
            self.domain, self.username = username.split('\\', 1)
        except ValueError:
            self.username = username
            self.domain = ''

        if self.domain:
            self.domain = self.domain.upper()
        self.password = password

        # This exposes the encrypt/decrypt methods used to encrypt and decrypt messages
        # sent after ntlm authentication. These methods are utilised by libraries that
        # call requests_ntlm to encrypt and decrypt the messages sent after authentication
        self.session_security = None

    def retry_using_http_NTLM_auth(self, auth_header_field, auth_header,
                                   response, auth_type, args):
        # Get the certificate of the server if using HTTPS for CBT
        server_certificate_hash = _get_server_cert(response)

        """Attempt to authenticate using HTTP NTLM challenge/response."""
        if auth_header in response.request.headers:
            return response

        content_length = int(
            response.request.headers.get('Content-Length', '0'), base=10)
        if hasattr(response.request.body, 'seek'):
            if content_length > 0:
                response.request.body.seek(-content_length, 1)
            else:
                response.request.body.seek(0, 0)

        # Consume content and release the original connection
        # to allow our new request to reuse the same one.
        response.content
        response.raw.release_conn()
        request = response.request.copy()

        # ntlm returns the headers as a base64 encoded bytestring. Convert to
        # a string.
        context = ntlm.Ntlm()
        negotiate_message = context.create_negotiate_message(self.domain).decode('ascii')
        auth = u'%s %s' % (auth_type, negotiate_message)
        request.headers[auth_header] = auth

        # A streaming response breaks authentication.
        # This can be fixed by not streaming this request, which is safe
        # because the returned response3 will still have stream=True set if
        # specified in args. In addition, we expect this request to give us a
        # challenge and not the real content, so the content will be short
        # anyway.
        args_nostream = dict(args, stream=False)
        response2 = response.connection.send(request, **args_nostream)

        # needed to make NTLM auth compatible with requests-2.3.0

        # Consume content and release the original connection
        # to allow our new request to reuse the same one.
        response2.content
        response2.raw.release_conn()
        request = response2.request.copy()

        # this is important for some web applications that store
        # authentication-related info in cookies (it took a long time to
        # figure out)
        if response2.headers.get('set-cookie'):
            request.headers['Cookie'] = response2.headers.get('set-cookie')

        # get the challenge
        auth_header_value = response2.headers[auth_header_field]

        auth_strip = auth_type + ' '

        ntlm_header_value = next(
            s for s in (val.lstrip() for val in auth_header_value.split(','))
            if s.startswith(auth_strip)
        ).strip()

        # Parse the challenge in the ntlm context
        context.parse_challenge_message(ntlm_header_value[len(auth_strip):])

        # build response
        # Get the response based on the challenge message
        authenticate_message = context.create_authenticate_message(
            self.username,
            self.password,
            self.domain,
            server_certificate_hash=server_certificate_hash
        )
        authenticate_message = authenticate_message.decode('ascii')
        auth = u'%s %s' % (auth_type, authenticate_message)
        request.headers[auth_header] = auth

        response3 = response2.connection.send(request, **args)

        # Update the history.
        response3.history.append(response)
        response3.history.append(response2)

        # Get the session_security object created by ntlm-auth for signing and sealing of messages
        self.session_security = context.session_security

        return response3

    def response_hook(self, r, **kwargs):
        """The actual hook handler."""
        if r.status_code == 401:
            # Handle server auth.
            www_authenticate = r.headers.get('www-authenticate', '').lower()
            auth_type = _auth_type_from_header(www_authenticate)

            if auth_type is not None:
                return self.retry_using_http_NTLM_auth(
                    'www-authenticate',
                    'Authorization',
                    r,
                    auth_type,
                    kwargs
                )
        elif r.status_code == 407:
            # If we didn't have server auth, do proxy auth.
            proxy_authenticate = r.headers.get(
                'proxy-authenticate', ''
            ).lower()
            auth_type = _auth_type_from_header(proxy_authenticate)
            if auth_type is not None:
                return self.retry_using_http_NTLM_auth(
                    'proxy-authenticate',
                    'Proxy-authorization',
                    r,
                    auth_type,
                    kwargs
                )

        return r

    def __call__(self, r):
        # we must keep the connection because NTLM authenticates the
        # connection, not single requests
        r.headers["Connection"] = "Keep-Alive"

        r.register_hook('response', self.response_hook)
        return r


def _auth_type_from_header(header):
    """
    Given a WWW-Authenticate or Proxy-Authenticate header, returns the
    authentication type to use. We prefer NTLM over Negotiate if the server
    suppports it.
    """
    if 'ntlm' in header:
        return 'NTLM'
    elif 'negotiate' in header:
        return 'Negotiate'

    return None

def _get_server_cert(response):
    """
    Get the certificate at the request_url and return it as a SHA256 hash. Will get the raw socket from the
    original response from the server. This socket is then checked if it is an SSL socket and then used to
    get the hash of the certificate. The certificate hash is then used with NTLMv2 authentication for
    Channel Binding Tokens support. If the raw object is not a urllib3 HTTPReponse (default with requests)
    then no certificate will be returned.

    :param response: The original 401 response from the server
    :return: SHA256 hash of the DER encoded certificate at the request_url or None if not a HTTPS endpoint
    """
    certificate_hash = None
    raw_response = response.raw

    if isinstance(raw_response, HTTPResponse):
        if sys.version_info > (3, 0):
            socket = raw_response._fp.fp.raw._sock
        else:
            socket = raw_response._fp.fp._sock

        try:
            server_certificate = socket.getpeercert(True)
        except AttributeError:
            pass
        else:
            hash_object = hashlib.sha256(server_certificate)
            certificate_hash = hash_object.hexdigest().upper()
    else:
        warnings.warn("Requests is running with a non urllib3 backend, cannot retrieve server certificate for CBT", NoCertificateRetrievedWarning)

    return certificate_hash

class NoCertificateRetrievedWarning(Warning):
    pass



from __future__ import with_statement
import os
import pytz
import subprocess
import sys

_cache_tz = None

if sys.version_info[0] == 2:

    class Popen(subprocess.Popen):

        def __enter__(self):
            return self

        def __exit__(self, type, value, traceback):
            if self.stdout:
                self.stdout.close()
            if self.stderr:
                self.stderr.close()
            if self.stdin:
                self.stdin.close()
            # Wait for the process to terminate, to avoid zombies.
            self.wait()

else:
    from subprocess import Popen


def _get_localzone(_root='/'):
    with Popen(
        "systemsetup -gettimezone",
        shell=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE
    ) as pipe:
        tzname = pipe.stdout.read().replace(b'Time Zone: ', b'').strip()

    if not tzname or tzname not in pytz.all_timezones_set:
        # link will be something like /usr/share/zoneinfo/America/Los_Angeles.
        link = os.readlink(os.path.join(_root, "etc/localtime"))
        tzname = link[link.rfind("zoneinfo/") + 9:]

    return pytz.timezone(tzname)


def get_localzone():
    """Get the computers configured local timezone, if any."""
    global _cache_tz
    if _cache_tz is None:
        _cache_tz = _get_localzone()
    return _cache_tz


def reload_localzone():
    """Reload the cached localzone. You need to call this if the timezone has changed."""
    global _cache_tz
    _cache_tz = _get_localzone()
    return _cache_tz



# -*- coding: utf-8 -*-
"""
    wakatime.language_priorities
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Overwrite pygments Lexer.priority attribute for specific languages.

    :copyright: (c) 2017 Alan Hamlett.
    :license: BSD, see LICENSE for more details.
"""


LANGUAGES = {
    'f#': 0.01,
    'perl': 0.01,
    'perl6': 0.01,
    'typescript': 0.01,
}



# -*- coding: utf-8 -*-
"""
    wakatime.configs
    ~~~~~~~~~~~~~~~~

    Config file parser.

    :copyright: (c) 2016 Alan Hamlett.
    :license: BSD, see LICENSE for more details.
"""


from __future__ import print_function

import os
import traceback

from .compat import open
from .constants import CONFIG_FILE_PARSE_ERROR


try:
    import configparser
except ImportError:  # pragma: nocover
    from .packages import configparser


def getConfigFile():
    """Returns the config file location.

    If $WAKATIME_HOME env varialbe is defined, returns
    $WAKATIME_HOME/.wakatime.cfg, otherwise ~/.wakatime.cfg.
    """

    fileName = '.wakatime.cfg'

    home = os.environ.get('WAKATIME_HOME')
    if home:
        return os.path.join(os.path.expanduser(home), fileName)

    return os.path.join(os.path.expanduser('~'), fileName)


def parseConfigFile(configFile=None):
    """Returns a configparser.SafeConfigParser instance with configs
    read from the config file. Default location of the config file is
    at ~/.wakatime.cfg.
    """

    # get config file location from ENV
    if not configFile:
        configFile = getConfigFile()

    configs = configparser.ConfigParser(delimiters=('='), strict=False)
    try:
        with open(configFile, 'r', encoding='utf-8') as fh:
            try:
                configs.read_file(fh)
            except configparser.Error:
                print(traceback.format_exc())
                raise SystemExit(CONFIG_FILE_PARSE_ERROR)
    except IOError:
        pass
    return configs



# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see <http://www.gnu.org/licenses/> or <http://www.gnu.org/licenses/lgpl.txt>.

"""
    [MS-NLMP] v28.0 2016-07-14

    2.2 Message Syntax
    The signature field used in NTLM messages
"""
NTLM_SIGNATURE = b'NTLMSSP\0'

"""
    [MS-NLMP] v28.0 2016-07-14

    2.2 Message Syntax
    The 3 message type options you can have in a message.
"""
class MessageTypes(object):
    NTLM_NEGOTIATE      = 0x1
    NTLM_CHALLENGE      = 0x2
    NTLM_AUTHENTICATE   = 0x3

"""
    [MS-NLMP] v28.0 2016-07-14

    2.2.2.1 AV_PAIR (MsvAvFlags)
    A 32-bit value indicated server or client configuration
"""
class AvFlags(object):
    AUTHENTICATION_CONSTRAINED  = 0x1
    MIC_PROVIDED                = 0x2
    UNTRUSTED_SPN_SOURCE        = 0x4

"""
    [MS-NLMP] v28.0 2016-07-14

    2.2.2.5 NEGOTIATE
    During NTLM authentication, each of the following flags is a possible value of the
    NegotiateFlags field of the NEGOTIATE_MESSAGE, CHALLENGE_MESSAGE and AUTHENTICATE_MESSAGE,
    unless otherwise noted. These flags define client or server NTLM capabilities
    supported by the sender.
"""
class NegotiateFlags(object):
    NTLMSSP_NEGOTIATE_56                            = 0x80000000
    NTLMSSP_NEGOTIATE_KEY_EXCH                      = 0x40000000
    NTLMSSP_NEGOTIATE_128                           = 0x20000000
    NTLMSSP_RESERVED_R1                             = 0x10000000
    NTLMSSP_RESERVED_R2                             = 0x08000000
    NTLMSSP_RESERVED_R3                             = 0x04000000
    NTLMSSP_NEGOTIATE_VERSION                       = 0x02000000
    NTLMSSP_RESERVED_R4                             = 0x01000000
    NTLMSSP_NEGOTIATE_TARGET_INFO                   = 0x00800000
    NTLMSSP_REQUEST_NON_NT_SESSION_KEY              = 0x00400000
    NTLMSSP_RESERVED_R5                             = 0x00200000
    NTLMSSP_NEGOTIATE_IDENTITY                      = 0x00100000
    NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY      = 0x00080000
    NTLMSSP_RESERVED_R6                             = 0x00040000
    NTLMSSP_TARGET_TYPE_SERVER                      = 0x00020000
    NTLMSSP_TARGET_TYPE_DOMAIN                      = 0x00010000
    NTLMSSP_NEGOTIATE_ALWAYS_SIGN                   = 0x00008000
    NTLMSSP_RESERVED_R7                             = 0x00004000
    NTLMSSP_NEGOTIATE_OEM_WORKSTATION_SUPPLIED      = 0x00002000
    NTLMSSP_NEGOTIATE_OEM_DOMAIN_SUPPLIED           = 0x00001000
    NTLMSSP_ANOYNMOUS                               = 0x00000800
    NTLMSSP_RESERVED_R8                             = 0x00000400
    NTLMSSP_NEGOTIATE_NTLM                          = 0x00000200
    NTLMSSP_RESERVED_R9                             = 0x00000100
    NTLMSSP_NEGOTIATE_LM_KEY                        = 0x00000080
    NTLMSSP_NEGOTIATE_DATAGRAM                      = 0x00000040
    NTLMSSP_NEGOTIATE_SEAL                          = 0x00000020
    NTLMSSP_NEGOTIATE_SIGN                          = 0x00000010
    NTLMSSP_RESERVED_R10                            = 0x00000008
    NTLMSSP_REQUEST_TARGET                          = 0x00000004
    NTLMSSP_NEGOTIATE_OEM                           = 0x00000002
    NTLMSSP_NEGOTIATE_UNICODE                       = 0x00000001

class SignSealConstants(object):
    # Magic Contants used to get the signing and sealing key for Extended Session Security
    CLIENT_SIGNING = b"session key to client-to-server signing key magic constant\0"
    SERVER_SIGNING = b"session key to server-to-client signing key magic constant\0"
    CLIENT_SEALING = b"session key to client-to-server sealing key magic constant\0"
    SERVER_SEALING = b"session key to server-to-client sealing key magic constant\0"



# -*- coding: utf-8 -*-
"""
    wakatime.projects.git
    ~~~~~~~~~~~~~~~~~~~~~

    Information about the git project for a given file.

    :copyright: (c) 2013 Alan Hamlett.
    :license: BSD, see LICENSE for more details.
"""

import logging
import os
import re
import sys

from .base import BaseProject
from ..compat import u, open


log = logging.getLogger('WakaTime')


class Git(BaseProject):
    _submodule = None
    _project_name = None
    _head_file = None

    def process(self):
        return self._find_git_config_file(self.path)

    def name(self):
        return u(self._project_name) if self._project_name else None

    def branch(self):
        head = self._head_file
        if head:
            try:
                with open(head, 'r', encoding='utf-8') as fh:
                    return self._get_branch_from_head_file(fh.readline())
            except UnicodeDecodeError:  # pragma: nocover
                try:
                    with open(head, 'r', encoding=sys.getfilesystemencoding()) as fh:
                        return self._get_branch_from_head_file(fh.readline())
                except:
                    log.traceback(logging.WARNING)
            except IOError:  # pragma: nocover
                log.traceback(logging.WARNING)
        return u('master')

    def _find_git_config_file(self, path):
        path = os.path.realpath(path)
        if os.path.isfile(path):
            path = os.path.split(path)[0]
        if os.path.isfile(os.path.join(path, '.git', 'config')):
            self._project_name = os.path.basename(path)
            self._head_file = os.path.join(path, '.git', 'HEAD')
            return True
        if self._submodules_supported_for_path(path):
            submodule_path = self._find_path_from_submodule(path)
            if submodule_path:
                self._project_name = os.path.basename(path)
                self._head_file = os.path.join(submodule_path, 'HEAD')
                return True
        split_path = os.path.split(path)
        if split_path[1] == '':
            return False
        return self._find_git_config_file(split_path[0])

    def _get_branch_from_head_file(self, line):
        if u(line.strip()).startswith('ref: '):
            return u(line.strip().rsplit('/', 1)[-1])
        return None

    def _submodules_supported_for_path(self, path):
        if not self._configs:
            return True

        disabled = self._configs.get('submodules_disabled')
        if not disabled:
            return True

        if disabled.strip().lower() == 'true':
            return False
        if disabled.strip().lower() == 'false':
            return True

        for pattern in disabled.split("\n"):
            if pattern.strip():
                try:
                    compiled = re.compile(pattern, re.IGNORECASE)
                    if compiled.search(path):
                        return False
                except re.error as ex:
                    log.warning(u('Regex error ({msg}) for disable git submodules pattern: {pattern}').format(
                        msg=u(ex),
                        pattern=u(pattern),
                    ))

        return True

    def _find_path_from_submodule(self, path):
        link = os.path.join(path, '.git')
        if not os.path.isfile(link):
            return None

        try:
            with open(link, 'r', encoding='utf-8') as fh:
                return self._get_path_from_submodule_link(path, fh.readline())
        except UnicodeDecodeError:
            try:
                with open(link, 'r', encoding=sys.getfilesystemencoding()) as fh:
                    return self._get_path_from_submodule_link(path, fh.readline())
            except:
                log.traceback(logging.WARNING)
        except IOError:
            log.traceback(logging.WARNING)

        return None

    def _get_path_from_submodule_link(self, path, line):
        if line.startswith('gitdir: '):
            subpath = line[len('gitdir: '):].strip()
            if os.path.isfile(os.path.join(path, subpath, 'config')) and \
                    os.path.isfile(os.path.join(path, subpath, 'HEAD')):
                return os.path.realpath(os.path.join(path, subpath))

        return None



# -*- coding: utf-8 -*-
"""
    wakatime.offlinequeue
    ~~~~~~~~~~~~~~~~~~~~~

    Queue for saving heartbeats while offline.

    :copyright: (c) 2014 Alan Hamlett.
    :license: BSD, see LICENSE for more details.
"""


import logging
import os
from time import sleep

from .compat import json
from .heartbeat import Heartbeat


try:
    import sqlite3
    HAS_SQL = True
except ImportError:  # pragma: nocover
    HAS_SQL = False


log = logging.getLogger('WakaTime')


class Queue(object):
    db_file = '.wakatime.db'
    table_name = 'heartbeat_2'

    args = None
    configs = None

    def __init__(self, args, configs):
        self.args = args
        self.configs = configs

    def connect(self):
        conn = sqlite3.connect(self._get_db_file(), isolation_level=None)
        c = conn.cursor()
        c.execute('''CREATE TABLE IF NOT EXISTS {0} (
            id text,
            heartbeat text)
        '''.format(self.table_name))
        return (conn, c)

    def push(self, heartbeat):
        if not HAS_SQL:
            return
        try:
            conn, c = self.connect()
            data = {
                'id': heartbeat.get_id(),
                'heartbeat': heartbeat.json(),
            }
            c.execute('INSERT INTO {0} VALUES (:id,:heartbeat)'.format(self.table_name), data)
            conn.commit()
            conn.close()
        except sqlite3.Error:
            log.traceback()

    def pop(self):
        if not HAS_SQL:
            return None
        tries = 3
        wait = 0.1
        try:
            conn, c = self.connect()
        except sqlite3.Error:
            log.traceback(logging.DEBUG)
            return None

        heartbeat = None

        loop = True
        while loop and tries > -1:
            try:
                c.execute('BEGIN IMMEDIATE')
                c.execute('SELECT * FROM {0} LIMIT 1'.format(self.table_name))
                row = c.fetchone()
                if row is not None:
                    id = row[0]
                    heartbeat = Heartbeat(json.loads(row[1]), self.args, self.configs, _clone=True)
                    c.execute('DELETE FROM {0} WHERE id=?'.format(self.table_name), [id])
                conn.commit()
                loop = False
            except sqlite3.Error:
                log.traceback(logging.DEBUG)
                sleep(wait)
                tries -= 1
        try:
            conn.close()
        except sqlite3.Error:
            log.traceback(logging.DEBUG)
        return heartbeat

    def push_many(self, heartbeats):
        for heartbeat in heartbeats:
            self.push(heartbeat)

    def pop_many(self, limit=None):
        if limit is None:
            limit = 5

        heartbeats = []

        count = 0
        while limit == 0 or count < limit:
            heartbeat = self.pop()
            if not heartbeat:
                break
            heartbeats.append(heartbeat)
            count += 1

        return heartbeats

    def _get_db_file(self):
        home = '~'
        if os.environ.get('WAKATIME_HOME'):
            home = os.environ.get('WAKATIME_HOME')
        return os.path.join(os.path.expanduser(home), '.wakatime.db')



# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see <http://www.gnu.org/licenses/> or <http://www.gnu.org/licenses/lgpl.txt>.

class ARC4(object):
    state = None
    i = 0
    j = 0

    def __init__(self, key):
        # Split up the key into a list
        if isinstance(key, str):
            key = [ord(c) for c in key]
        else:
            key = [c for c in key]

        #Key-scheduling algorithm (KSA)
        self.state = [n for n in range(256)]
        j = 0
        for i in range(256):
            j = (j + self.state[i] + key[i % len(key)]) % 256
            self.state[i], self.state[j] = self.state[j], self.state[i]

    def update(self, value):
        chars = []
        random_gen = self._random_generator()
        for char in value:
            if isinstance(value, str):
                byte = ord(char)
            else:
                byte = char
            updated_byte = byte ^ next(random_gen)
            chars.append(updated_byte)
        return bytes(bytearray(chars))

    def _random_generator(self):
        #Pseudo-Random Generation Algorithm (PRGA)
        while True:
            self.i = (self.i + 1) % 256
            self.j = (self.j + self.state[self.i]) % 256
            self.state[self.i], self.state[self.j] = self.state[self.j], self.state[self.i]
            yield self.state[(self.state[self.i] + self.state[self.j]) % 256]



# -*- coding: utf-8 -*-
"""
    pygments.lexers.haskell
    ~~~~~~~~~~~~~~~~~~~~~~~

    Lexers for Haskell and related languages.

    :copyright: Copyright 2006-2017 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

import re

from pygments.lexer import Lexer, RegexLexer, bygroups, do_insertions, \
    default, include
from pygments.token import Text, Comment, Operator, Keyword, Name, String, \
    Number, Punctuation, Generic
from pygments import unistring as uni

__all__ = ['HaskellLexer', 'IdrisLexer', 'AgdaLexer', 'CryptolLexer',
           'LiterateHaskellLexer', 'LiterateIdrisLexer', 'LiterateAgdaLexer',
           'LiterateCryptolLexer', 'KokaLexer']


line_re = re.compile('.*?\n')


class HaskellLexer(RegexLexer):
    """
    A Haskell lexer based on the lexemes defined in the Haskell 98 Report.

    .. versionadded:: 0.8
    """
    name = 'Haskell'
    aliases = ['haskell', 'hs']
    filenames = ['*.hs']
    mimetypes = ['text/x-haskell']

    flags = re.MULTILINE | re.UNICODE

    reserved = ('case', 'class', 'data', 'default', 'deriving', 'do', 'else',
                'family', 'if', 'in', 'infix[lr]?', 'instance',
                'let', 'newtype', 'of', 'then', 'type', 'where', '_')
    ascii = ('NUL', 'SOH', '[SE]TX', 'EOT', 'ENQ', 'ACK',
             'BEL', 'BS', 'HT', 'LF', 'VT', 'FF', 'CR', 'S[OI]', 'DLE',
             'DC[1-4]', 'NAK', 'SYN', 'ETB', 'CAN',
             'EM', 'SUB', 'ESC', '[FGRU]S', 'SP', 'DEL')

    tokens = {
        'root': [
            # Whitespace:
            (r'\s+', Text),
            # (r'--\s*|.*$', Comment.Doc),
            (r'--(?![!#$%&*+./<=>?@^|_~:\\]).*?$', Comment.Single),
            (r'\{-', Comment.Multiline, 'comment'),
            # Lexemes:
            #  Identifiers
            (r'\bimport\b', Keyword.Reserved, 'import'),
            (r'\bmodule\b', Keyword.Reserved, 'module'),
            (r'\berror\b', Name.Exception),
            (r'\b(%s)(?!\')\b' % '|'.join(reserved), Keyword.Reserved),
            (r"'[^\\]'", String.Char),  # this has to come before the TH quote
            (r'^[_' + uni.Ll + r'][\w\']*', Name.Function),
            (r"'?[_" + uni.Ll + r"][\w']*", Name),
            (r"('')?[" + uni.Lu + r"][\w\']*", Keyword.Type),
            (r"(')[" + uni.Lu + r"][\w\']*", Keyword.Type),
            (r"(')\[[^\]]*\]", Keyword.Type),  # tuples and lists get special treatment in GHC
            (r"(')\([^)]*\)", Keyword.Type),  # ..
            #  Operators
            (r'\\(?![:!#$%&*+.\\/<=>?@^|~-]+)', Name.Function),  # lambda operator
            (r'(<-|::|->|=>|=)(?![:!#$%&*+.\\/<=>?@^|~-]+)', Operator.Word),  # specials
            (r':[:!#$%&*+.\\/<=>?@^|~-]*', Keyword.Type),  # Constructor operators
            (r'[:!#$%&*+.\\/<=>?@^|~-]+', Operator),  # Other operators
            #  Numbers
            (r'\d+[eE][+-]?\d+', Number.Float),
            (r'\d+\.\d+([eE][+-]?\d+)?', Number.Float),
            (r'0[oO][0-7]+', Number.Oct),
            (r'0[xX][\da-fA-F]+', Number.Hex),
            (r'\d+', Number.Integer),
            #  Character/String Literals
            (r"'", String.Char, 'character'),
            (r'"', String, 'string'),
            #  Special
            (r'\[\]', Keyword.Type),
            (r'\(\)', Name.Builtin),
            (r'[][(),;`{}]', Punctuation),
        ],
        'import': [
            # Import statements
            (r'\s+', Text),
            (r'"', String, 'string'),
            # after "funclist" state
            (r'\)', Punctuation, '#pop'),
            (r'qualified\b', Keyword),
            # import X as Y
            (r'([' + uni.Lu + r'][\w.]*)(\s+)(as)(\s+)([' + uni.Lu + r'][\w.]*)',
             bygroups(Name.Namespace, Text, Keyword, Text, Name), '#pop'),
            # import X hiding (functions)
            (r'([' + uni.Lu + r'][\w.]*)(\s+)(hiding)(\s+)(\()',
             bygroups(Name.Namespace, Text, Keyword, Text, Punctuation), 'funclist'),
            # import X (functions)
            (r'([' + uni.Lu + r'][\w.]*)(\s+)(\()',
             bygroups(Name.Namespace, Text, Punctuation), 'funclist'),
            # import X
            (r'[\w.]+', Name.Namespace, '#pop'),
        ],
        'module': [
            (r'\s+', Text),
            (r'([' + uni.Lu + r'][\w.]*)(\s+)(\()',
             bygroups(Name.Namespace, Text, Punctuation), 'funclist'),
            (r'[' + uni.Lu + r'][\w.]*', Name.Namespace, '#pop'),
        ],
        'funclist': [
            (r'\s+', Text),
            (r'[' + uni.Lu + r']\w*', Keyword.Type),
            (r'(_[\w\']+|[' + uni.Ll + r'][\w\']*)', Name.Function),
            (r'--(?![!#$%&*+./<=>?@^|_~:\\]).*?$', Comment.Single),
            (r'\{-', Comment.Multiline, 'comment'),
            (r',', Punctuation),
            (r'[:!#$%&*+.\\/<=>?@^|~-]+', Operator),
            # (HACK, but it makes sense to push two instances, believe me)
            (r'\(', Punctuation, ('funclist', 'funclist')),
            (r'\)', Punctuation, '#pop:2'),
        ],
        # NOTE: the next four states are shared in the AgdaLexer; make sure
        # any change is compatible with Agda as well or copy over and change
        'comment': [
            # Multiline Comments
            (r'[^-{}]+', Comment.Multiline),
            (r'\{-', Comment.Multiline, '#push'),
            (r'-\}', Comment.Multiline, '#pop'),
            (r'[-{}]', Comment.Multiline),
        ],
        'character': [
            # Allows multi-chars, incorrectly.
            (r"[^\\']'", String.Char, '#pop'),
            (r"\\", String.Escape, 'escape'),
            ("'", String.Char, '#pop'),
        ],
        'string': [
            (r'[^\\"]+', String),
            (r"\\", String.Escape, 'escape'),
            ('"', String, '#pop'),
        ],
        'escape': [
            (r'[abfnrtv"\'&\\]', String.Escape, '#pop'),
            (r'\^[][' + uni.Lu + r'@^_]', String.Escape, '#pop'),
            ('|'.join(ascii), String.Escape, '#pop'),
            (r'o[0-7]+', String.Escape, '#pop'),
            (r'x[\da-fA-F]+', String.Escape, '#pop'),
            (r'\d+', String.Escape, '#pop'),
            (r'\s+\\', String.Escape, '#pop'),
        ],
    }


class IdrisLexer(RegexLexer):
    """
    A lexer for the dependently typed programming language Idris.

    Based on the Haskell and Agda Lexer.

    .. versionadded:: 2.0
    """
    name = 'Idris'
    aliases = ['idris', 'idr']
    filenames = ['*.idr']
    mimetypes = ['text/x-idris']

    reserved = ('case', 'class', 'data', 'default', 'using', 'do', 'else',
                'if', 'in', 'infix[lr]?', 'instance', 'rewrite', 'auto',
                'namespace', 'codata', 'mutual', 'private', 'public', 'abstract',
                'total', 'partial',
                'let', 'proof', 'of', 'then', 'static', 'where', '_', 'with',
                'pattern',  'term',  'syntax', 'prefix',
                'postulate', 'parameters', 'record', 'dsl', 'impossible', 'implicit',
                'tactics', 'intros', 'intro', 'compute', 'refine', 'exact', 'trivial')

    ascii = ('NUL', 'SOH', '[SE]TX', 'EOT', 'ENQ', 'ACK',
             'BEL', 'BS', 'HT', 'LF', 'VT', 'FF', 'CR', 'S[OI]', 'DLE',
             'DC[1-4]', 'NAK', 'SYN', 'ETB', 'CAN',
             'EM', 'SUB', 'ESC', '[FGRU]S', 'SP', 'DEL')

    directives = ('lib', 'link', 'flag', 'include', 'hide', 'freeze', 'access',
                  'default', 'logging', 'dynamic', 'name', 'error_handlers', 'language')

    tokens = {
        'root': [
            # Comments
            (r'^(\s*)(%%%s)' % '|'.join(directives),
             bygroups(Text, Keyword.Reserved)),
            (r'(\s*)(--(?![!#$%&*+./<=>?@^|_~:\\]).*?)$', bygroups(Text, Comment.Single)),
            (r'(\s*)(\|{3}.*?)$', bygroups(Text, Comment.Single)),
            (r'(\s*)(\{-)', bygroups(Text, Comment.Multiline), 'comment'),
            # Declaration
            (r'^(\s*)([^\s(){}]+)(\s*)(:)(\s*)',
             bygroups(Text, Name.Function, Text, Operator.Word, Text)),
            #  Identifiers
            (r'\b(%s)(?!\')\b' % '|'.join(reserved), Keyword.Reserved),
            (r'(import|module)(\s+)', bygroups(Keyword.Reserved, Text), 'module'),
            (r"('')?[A-Z][\w\']*", Keyword.Type),
            (r'[a-z][\w\']*', Text),
            #  Special Symbols
            (r'(<-|::|->|=>|=)', Operator.Word),  # specials
            (r'([(){}\[\]:!#$%&*+.\\/<=>?@^|~-]+)', Operator.Word),  # specials
            #  Numbers
            (r'\d+[eE][+-]?\d+', Number.Float),
            (r'\d+\.\d+([eE][+-]?\d+)?', Number.Float),
            (r'0[xX][\da-fA-F]+', Number.Hex),
            (r'\d+', Number.Integer),
            # Strings
            (r"'", String.Char, 'character'),
            (r'"', String, 'string'),
            (r'[^\s(){}]+', Text),
            (r'\s+?', Text),  # Whitespace
        ],
        'module': [
            (r'\s+', Text),
            (r'([A-Z][\w.]*)(\s+)(\()',
             bygroups(Name.Namespace, Text, Punctuation), 'funclist'),
            (r'[A-Z][\w.]*', Name.Namespace, '#pop'),
        ],
        'funclist': [
            (r'\s+', Text),
            (r'[A-Z]\w*', Keyword.Type),
            (r'(_[\w\']+|[a-z][\w\']*)', Name.Function),
            (r'--.*$', Comment.Single),
            (r'\{-', Comment.Multiline, 'comment'),
            (r',', Punctuation),
            (r'[:!#$%&*+.\\/<=>?@^|~-]+', Operator),
            # (HACK, but it makes sense to push two instances, believe me)
            (r'\(', Punctuation, ('funclist', 'funclist')),
            (r'\)', Punctuation, '#pop:2'),
        ],
        # NOTE: the next four states are shared in the AgdaLexer; make sure
        # any change is compatible with Agda as well or copy over and change
        'comment': [
            # Multiline Comments
            (r'[^-{}]+', Comment.Multiline),
            (r'\{-', Comment.Multiline, '#push'),
            (r'-\}', Comment.Multiline, '#pop'),
            (r'[-{}]', Comment.Multiline),
        ],
        'character': [
            # Allows multi-chars, incorrectly.
            (r"[^\\']", String.Char),
            (r"\\", String.Escape, 'escape'),
            ("'", String.Char, '#pop'),
        ],
        'string': [
            (r'[^\\"]+', String),
            (r"\\", String.Escape, 'escape'),
            ('"', String, '#pop'),
        ],
        'escape': [
            (r'[abfnrtv"\'&\\]', String.Escape, '#pop'),
            (r'\^[][A-Z@^_]', String.Escape, '#pop'),
            ('|'.join(ascii), String.Escape, '#pop'),
            (r'o[0-7]+', String.Escape, '#pop'),
            (r'x[\da-fA-F]+', String.Escape, '#pop'),
            (r'\d+', String.Escape, '#pop'),
            (r'\s+\\', String.Escape, '#pop')
        ],
    }


class AgdaLexer(RegexLexer):
    """
    For the `Agda <http://wiki.portal.chalmers.se/agda/pmwiki.php>`_
    dependently typed functional programming language and proof assistant.

    .. versionadded:: 2.0
    """

    name = 'Agda'
    aliases = ['agda']
    filenames = ['*.agda']
    mimetypes = ['text/x-agda']

    reserved = ['abstract', 'codata', 'coinductive', 'constructor', 'data',
                'field', 'forall', 'hiding', 'in', 'inductive', 'infix',
                'infixl', 'infixr', 'instance', 'let', 'mutual', 'open',
                'pattern', 'postulate', 'primitive', 'private',
                'quote', 'quoteGoal', 'quoteTerm',
                'record', 'renaming', 'rewrite', 'syntax', 'tactic',
                'unquote', 'unquoteDecl', 'using', 'where', 'with']

    tokens = {
        'root': [
            # Declaration
            (r'^(\s*)([^\s(){}]+)(\s*)(:)(\s*)',
             bygroups(Text, Name.Function, Text, Operator.Word, Text)),
            # Comments
            (r'--(?![!#$%&*+./<=>?@^|_~:\\]).*?$', Comment.Single),
            (r'\{-', Comment.Multiline, 'comment'),
            # Holes
            (r'\{!', Comment.Directive, 'hole'),
            # Lexemes:
            #  Identifiers
            (r'\b(%s)(?!\')\b' % '|'.join(reserved), Keyword.Reserved),
            (r'(import|module)(\s+)', bygroups(Keyword.Reserved, Text), 'module'),
            (r'\b(Set|Prop)\b', Keyword.Type),
            #  Special Symbols
            (r'(\(|\)|\{|\})', Operator),
            (u'(\\.{1,3}|\\||\u039B|\u2200|\u2192|:|=|->)', Operator.Word),
            #  Numbers
            (r'\d+[eE][+-]?\d+', Number.Float),
            (r'\d+\.\d+([eE][+-]?\d+)?', Number.Float),
            (r'0[xX][\da-fA-F]+', Number.Hex),
            (r'\d+', Number.Integer),
            # Strings
            (r"'", String.Char, 'character'),
            (r'"', String, 'string'),
            (r'[^\s(){}]+', Text),
            (r'\s+?', Text),  # Whitespace
        ],
        'hole': [
            # Holes
            (r'[^!{}]+', Comment.Directive),
            (r'\{!', Comment.Directive, '#push'),
            (r'!\}', Comment.Directive, '#pop'),
            (r'[!{}]', Comment.Directive),
        ],
        'module': [
            (r'\{-', Comment.Multiline, 'comment'),
            (r'[a-zA-Z][\w.]*', Name, '#pop'),
            (r'[\W0-9_]+', Text)
        ],
        'comment': HaskellLexer.tokens['comment'],
        'character': HaskellLexer.tokens['character'],
        'string': HaskellLexer.tokens['string'],
        'escape': HaskellLexer.tokens['escape']
    }


class CryptolLexer(RegexLexer):
    """
    FIXME: A Cryptol2 lexer based on the lexemes defined in the Haskell 98 Report.

    .. versionadded:: 2.0
    """
    name = 'Cryptol'
    aliases = ['cryptol', 'cry']
    filenames = ['*.cry']
    mimetypes = ['text/x-cryptol']

    reserved = ('Arith', 'Bit', 'Cmp', 'False', 'Inf', 'True', 'else',
                'export', 'extern', 'fin', 'if', 'import', 'inf', 'lg2',
                'max', 'min', 'module', 'newtype', 'pragma', 'property',
                'then', 'type', 'where', 'width')
    ascii = ('NUL', 'SOH', '[SE]TX', 'EOT', 'ENQ', 'ACK',
             'BEL', 'BS', 'HT', 'LF', 'VT', 'FF', 'CR', 'S[OI]', 'DLE',
             'DC[1-4]', 'NAK', 'SYN', 'ETB', 'CAN',
             'EM', 'SUB', 'ESC', '[FGRU]S', 'SP', 'DEL')

    tokens = {
        'root': [
            # Whitespace:
            (r'\s+', Text),
            # (r'--\s*|.*$', Comment.Doc),
            (r'//.*$', Comment.Single),
            (r'/\*', Comment.Multiline, 'comment'),
            # Lexemes:
            #  Identifiers
            (r'\bimport\b', Keyword.Reserved, 'import'),
            (r'\bmodule\b', Keyword.Reserved, 'module'),
            (r'\berror\b', Name.Exception),
            (r'\b(%s)(?!\')\b' % '|'.join(reserved), Keyword.Reserved),
            (r'^[_a-z][\w\']*', Name.Function),
            (r"'?[_a-z][\w']*", Name),
            (r"('')?[A-Z][\w\']*", Keyword.Type),
            #  Operators
            (r'\\(?![:!#$%&*+.\\/<=>?@^|~-]+)', Name.Function),  # lambda operator
            (r'(<-|::|->|=>|=)(?![:!#$%&*+.\\/<=>?@^|~-]+)', Operator.Word),  # specials
            (r':[:!#$%&*+.\\/<=>?@^|~-]*', Keyword.Type),  # Constructor operators
            (r'[:!#$%&*+.\\/<=>?@^|~-]+', Operator),  # Other operators
            #  Numbers
            (r'\d+[eE][+-]?\d+', Number.Float),
            (r'\d+\.\d+([eE][+-]?\d+)?', Number.Float),
            (r'0[oO][0-7]+', Number.Oct),
            (r'0[xX][\da-fA-F]+', Number.Hex),
            (r'\d+', Number.Integer),
            #  Character/String Literals
            (r"'", String.Char, 'character'),
            (r'"', String, 'string'),
            #  Special
            (r'\[\]', Keyword.Type),
            (r'\(\)', Name.Builtin),
            (r'[][(),;`{}]', Punctuation),
        ],
        'import': [
            # Import statements
            (r'\s+', Text),
            (r'"', String, 'string'),
            # after "funclist" state
            (r'\)', Punctuation, '#pop'),
            (r'qualified\b', Keyword),
            # import X as Y
            (r'([A-Z][\w.]*)(\s+)(as)(\s+)([A-Z][\w.]*)',
             bygroups(Name.Namespace, Text, Keyword, Text, Name), '#pop'),
            # import X hiding (functions)
            (r'([A-Z][\w.]*)(\s+)(hiding)(\s+)(\()',
             bygroups(Name.Namespace, Text, Keyword, Text, Punctuation), 'funclist'),
            # import X (functions)
            (r'([A-Z][\w.]*)(\s+)(\()',
             bygroups(Name.Namespace, Text, Punctuation), 'funclist'),
            # import X
            (r'[\w.]+', Name.Namespace, '#pop'),
        ],
        'module': [
            (r'\s+', Text),
            (r'([A-Z][\w.]*)(\s+)(\()',
             bygroups(Name.Namespace, Text, Punctuation), 'funclist'),
            (r'[A-Z][\w.]*', Name.Namespace, '#pop'),
        ],
        'funclist': [
            (r'\s+', Text),
            (r'[A-Z]\w*', Keyword.Type),
            (r'(_[\w\']+|[a-z][\w\']*)', Name.Function),
            # TODO: these don't match the comments in docs, remove.
            #(r'--(?![!#$%&*+./<=>?@^|_~:\\]).*?$', Comment.Single),
            #(r'{-', Comment.Multiline, 'comment'),
            (r',', Punctuation),
            (r'[:!#$%&*+.\\/<=>?@^|~-]+', Operator),
            # (HACK, but it makes sense to push two instances, believe me)
            (r'\(', Punctuation, ('funclist', 'funclist')),
            (r'\)', Punctuation, '#pop:2'),
        ],
        'comment': [
            # Multiline Comments
            (r'[^/*]+', Comment.Multiline),
            (r'/\*', Comment.Multiline, '#push'),
            (r'\*/', Comment.Multiline, '#pop'),
            (r'[*/]', Comment.Multiline),
        ],
        'character': [
            # Allows multi-chars, incorrectly.
            (r"[^\\']'", String.Char, '#pop'),
            (r"\\", String.Escape, 'escape'),
            ("'", String.Char, '#pop'),
        ],
        'string': [
            (r'[^\\"]+', String),
            (r"\\", String.Escape, 'escape'),
            ('"', String, '#pop'),
        ],
        'escape': [
            (r'[abfnrtv"\'&\\]', String.Escape, '#pop'),
            (r'\^[][A-Z@^_]', String.Escape, '#pop'),
            ('|'.join(ascii), String.Escape, '#pop'),
            (r'o[0-7]+', String.Escape, '#pop'),
            (r'x[\da-fA-F]+', String.Escape, '#pop'),
            (r'\d+', String.Escape, '#pop'),
            (r'\s+\\', String.Escape, '#pop'),
        ],
    }

    EXTRA_KEYWORDS = set(('join', 'split', 'reverse', 'transpose', 'width',
                          'length', 'tail', '<<', '>>', '<<<', '>>>', 'const',
                          'reg', 'par', 'seq', 'ASSERT', 'undefined', 'error',
                          'trace'))

    def get_tokens_unprocessed(self, text):
        stack = ['root']
        for index, token, value in \
                RegexLexer.get_tokens_unprocessed(self, text, stack):
            if token is Name and value in self.EXTRA_KEYWORDS:
                yield index, Name.Builtin, value
            else:
                yield index, token, value


class LiterateLexer(Lexer):
    """
    Base class for lexers of literate file formats based on LaTeX or Bird-style
    (prefixing each code line with ">").

    Additional options accepted:

    `litstyle`
        If given, must be ``"bird"`` or ``"latex"``.  If not given, the style
        is autodetected: if the first non-whitespace character in the source
        is a backslash or percent character, LaTeX is assumed, else Bird.
    """

    bird_re = re.compile(r'(>[ \t]*)(.*\n)')

    def __init__(self, baselexer, **options):
        self.baselexer = baselexer
        Lexer.__init__(self, **options)

    def get_tokens_unprocessed(self, text):
        style = self.options.get('litstyle')
        if style is None:
            style = (text.lstrip()[0:1] in '%\\') and 'latex' or 'bird'

        code = ''
        insertions = []
        if style == 'bird':
            # bird-style
            for match in line_re.finditer(text):
                line = match.group()
                m = self.bird_re.match(line)
                if m:
                    insertions.append((len(code),
                                       [(0, Comment.Special, m.group(1))]))
                    code += m.group(2)
                else:
                    insertions.append((len(code), [(0, Text, line)]))
        else:
            # latex-style
            from pygments.lexers.markup import TexLexer
            lxlexer = TexLexer(**self.options)
            codelines = 0
            latex = ''
            for match in line_re.finditer(text):
                line = match.group()
                if codelines:
                    if line.lstrip().startswith('\\end{code}'):
                        codelines = 0
                        latex += line
                    else:
                        code += line
                elif line.lstrip().startswith('\\begin{code}'):
                    codelines = 1
                    latex += line
                    insertions.append((len(code),
                                       list(lxlexer.get_tokens_unprocessed(latex))))
                    latex = ''
                else:
                    latex += line
            insertions.append((len(code),
                               list(lxlexer.get_tokens_unprocessed(latex))))
        for item in do_insertions(insertions, self.baselexer.get_tokens_unprocessed(code)):
            yield item


class LiterateHaskellLexer(LiterateLexer):
    """
    For Literate Haskell (Bird-style or LaTeX) source.

    Additional options accepted:

    `litstyle`
        If given, must be ``"bird"`` or ``"latex"``.  If not given, the style
        is autodetected: if the first non-whitespace character in the source
        is a backslash or percent character, LaTeX is assumed, else Bird.

    .. versionadded:: 0.9
    """
    name = 'Literate Haskell'
    aliases = ['lhs', 'literate-haskell', 'lhaskell']
    filenames = ['*.lhs']
    mimetypes = ['text/x-literate-haskell']

    def __init__(self, **options):
        hslexer = HaskellLexer(**options)
        LiterateLexer.__init__(self, hslexer, **options)


class LiterateIdrisLexer(LiterateLexer):
    """
    For Literate Idris (Bird-style or LaTeX) source.

    Additional options accepted:

    `litstyle`
        If given, must be ``"bird"`` or ``"latex"``.  If not given, the style
        is autodetected: if the first non-whitespace character in the source
        is a backslash or percent character, LaTeX is assumed, else Bird.

    .. versionadded:: 2.0
    """
    name = 'Literate Idris'
    aliases = ['lidr', 'literate-idris', 'lidris']
    filenames = ['*.lidr']
    mimetypes = ['text/x-literate-idris']

    def __init__(self, **options):
        hslexer = IdrisLexer(**options)
        LiterateLexer.__init__(self, hslexer, **options)


class LiterateAgdaLexer(LiterateLexer):
    """
    For Literate Agda source.

    Additional options accepted:

    `litstyle`
        If given, must be ``"bird"`` or ``"latex"``.  If not given, the style
        is autodetected: if the first non-whitespace character in the source
        is a backslash or percent character, LaTeX is assumed, else Bird.

    .. versionadded:: 2.0
    """
    name = 'Literate Agda'
    aliases = ['lagda', 'literate-agda']
    filenames = ['*.lagda']
    mimetypes = ['text/x-literate-agda']

    def __init__(self, **options):
        agdalexer = AgdaLexer(**options)
        LiterateLexer.__init__(self, agdalexer, litstyle='latex', **options)


class LiterateCryptolLexer(LiterateLexer):
    """
    For Literate Cryptol (Bird-style or LaTeX) source.

    Additional options accepted:

    `litstyle`
        If given, must be ``"bird"`` or ``"latex"``.  If not given, the style
        is autodetected: if the first non-whitespace character in the source
        is a backslash or percent character, LaTeX is assumed, else Bird.

    .. versionadded:: 2.0
    """
    name = 'Literate Cryptol'
    aliases = ['lcry', 'literate-cryptol', 'lcryptol']
    filenames = ['*.lcry']
    mimetypes = ['text/x-literate-cryptol']

    def __init__(self, **options):
        crylexer = CryptolLexer(**options)
        LiterateLexer.__init__(self, crylexer, **options)


class KokaLexer(RegexLexer):
    """
    Lexer for the `Koka <http://koka.codeplex.com>`_
    language.

    .. versionadded:: 1.6
    """

    name = 'Koka'
    aliases = ['koka']
    filenames = ['*.kk', '*.kki']
    mimetypes = ['text/x-koka']

    keywords = [
        'infix', 'infixr', 'infixl',
        'type', 'cotype', 'rectype', 'alias',
        'struct', 'con',
        'fun', 'function', 'val', 'var',
        'external',
        'if', 'then', 'else', 'elif', 'return', 'match',
        'private', 'public', 'private',
        'module', 'import', 'as',
        'include', 'inline',
        'rec',
        'try', 'yield', 'enum',
        'interface', 'instance',
    ]

    # keywords that are followed by a type
    typeStartKeywords = [
        'type', 'cotype', 'rectype', 'alias', 'struct', 'enum',
    ]

    # keywords valid in a type
    typekeywords = [
        'forall', 'exists', 'some', 'with',
    ]

    # builtin names and special names
    builtin = [
        'for', 'while', 'repeat',
        'foreach', 'foreach-indexed',
        'error', 'catch', 'finally',
        'cs', 'js', 'file', 'ref', 'assigned',
    ]

    # symbols that can be in an operator
    symbols = r'[$%&*+@!/\\^~=.:\-?|<>]+'

    # symbol boundary: an operator keyword should not be followed by any of these
    sboundary = '(?!'+symbols+')'

    # name boundary: a keyword should not be followed by any of these
    boundary = '(?![\w/])'

    # koka token abstractions
    tokenType = Name.Attribute
    tokenTypeDef = Name.Class
    tokenConstructor = Generic.Emph

    # main lexer
    tokens = {
        'root': [
            include('whitespace'),

            # go into type mode
            (r'::?' + sboundary, tokenType, 'type'),
            (r'(alias)(\s+)([a-z]\w*)?', bygroups(Keyword, Text, tokenTypeDef),
             'alias-type'),
            (r'(struct)(\s+)([a-z]\w*)?', bygroups(Keyword, Text, tokenTypeDef),
             'struct-type'),
            ((r'(%s)' % '|'.join(typeStartKeywords)) +
             r'(\s+)([a-z]\w*)?', bygroups(Keyword, Text, tokenTypeDef),
             'type'),

            # special sequences of tokens (we use ?: for non-capturing group as
            # required by 'bygroups')
            (r'(module)(\s+)(interface\s+)?((?:[a-z]\w*/)*[a-z]\w*)',
             bygroups(Keyword, Text, Keyword, Name.Namespace)),
            (r'(import)(\s+)((?:[a-z]\w*/)*[a-z]\w*)'
             r'(?:(\s*)(=)(\s*)((?:qualified\s*)?)'
             r'((?:[a-z]\w*/)*[a-z]\w*))?',
             bygroups(Keyword, Text, Name.Namespace, Text, Keyword, Text,
                      Keyword, Name.Namespace)),

            (r'(^(?:(?:public|private)\s*)?(?:function|fun|val))'
             r'(\s+)([a-z]\w*|\((?:' + symbols + r'|/)\))',
             bygroups(Keyword, Text, Name.Function)),
            (r'(^(?:(?:public|private)\s*)?external)(\s+)(inline\s+)?'
             r'([a-z]\w*|\((?:' + symbols + r'|/)\))',
             bygroups(Keyword, Text, Keyword, Name.Function)),

            # keywords
            (r'(%s)' % '|'.join(typekeywords) + boundary, Keyword.Type),
            (r'(%s)' % '|'.join(keywords) + boundary, Keyword),
            (r'(%s)' % '|'.join(builtin) + boundary, Keyword.Pseudo),
            (r'::?|:=|\->|[=.]' + sboundary, Keyword),

            # names
            (r'((?:[a-z]\w*/)*)([A-Z]\w*)',
             bygroups(Name.Namespace, tokenConstructor)),
            (r'((?:[a-z]\w*/)*)([a-z]\w*)', bygroups(Name.Namespace, Name)),
            (r'((?:[a-z]\w*/)*)(\((?:' + symbols + r'|/)\))',
             bygroups(Name.Namespace, Name)),
            (r'_\w*', Name.Variable),

            # literal string
            (r'@"', String.Double, 'litstring'),

            # operators
            (symbols + "|/(?![*/])", Operator),
            (r'`', Operator),
            (r'[{}()\[\];,]', Punctuation),

            # literals. No check for literal characters with len > 1
            (r'[0-9]+\.[0-9]+([eE][\-+]?[0-9]+)?', Number.Float),
            (r'0[xX][0-9a-fA-F]+', Number.Hex),
            (r'[0-9]+', Number.Integer),

            (r"'", String.Char, 'char'),
            (r'"', String.Double, 'string'),
        ],

        # type started by alias
        'alias-type': [
            (r'=', Keyword),
            include('type')
        ],

        # type started by struct
        'struct-type': [
            (r'(?=\((?!,*\)))', Punctuation, '#pop'),
            include('type')
        ],

        # type started by colon
        'type': [
            (r'[(\[<]', tokenType, 'type-nested'),
            include('type-content')
        ],

        # type nested in brackets: can contain parameters, comma etc.
        'type-nested': [
            (r'[)\]>]', tokenType, '#pop'),
            (r'[(\[<]', tokenType, 'type-nested'),
            (r',', tokenType),
            (r'([a-z]\w*)(\s*)(:)(?!:)',
             bygroups(Name, Text, tokenType)),  # parameter name
            include('type-content')
        ],

        # shared contents of a type
        'type-content': [
            include('whitespace'),

            # keywords
            (r'(%s)' % '|'.join(typekeywords) + boundary, Keyword),
            (r'(?=((%s)' % '|'.join(keywords) + boundary + '))',
             Keyword, '#pop'),  # need to match because names overlap...

            # kinds
            (r'[EPHVX]' + boundary, tokenType),

            # type names
            (r'[a-z][0-9]*(?![\w/])', tokenType),
            (r'_\w*', tokenType.Variable),  # Generic.Emph
            (r'((?:[a-z]\w*/)*)([A-Z]\w*)',
             bygroups(Name.Namespace, tokenType)),
            (r'((?:[a-z]\w*/)*)([a-z]\w+)',
             bygroups(Name.Namespace, tokenType)),

            # type keyword operators
            (r'::|->|[.:|]', tokenType),

            # catchall
            default('#pop')
        ],

        # comments and literals
        'whitespace': [
            (r'\n\s*#.*$', Comment.Preproc),
            (r'\s+', Text),
            (r'/\*', Comment.Multiline, 'comment'),
            (r'//.*$', Comment.Single)
        ],
        'comment': [
            (r'[^/*]+', Comment.Multiline),
            (r'/\*', Comment.Multiline, '#push'),
            (r'\*/', Comment.Multiline, '#pop'),
            (r'[*/]', Comment.Multiline),
        ],
        'litstring': [
            (r'[^"]+', String.Double),
            (r'""', String.Escape),
            (r'"', String.Double, '#pop'),
        ],
        'string': [
            (r'[^\\"\n]+', String.Double),
            include('escape-sequence'),
            (r'["\n]', String.Double, '#pop'),
        ],
        'char': [
            (r'[^\\\'\n]+', String.Char),
            include('escape-sequence'),
            (r'[\'\n]', String.Char, '#pop'),
        ],
        'escape-sequence': [
            (r'\\[nrt\\"\']', String.Escape),
            (r'\\x[0-9a-fA-F]{2}', String.Escape),
            (r'\\u[0-9a-fA-F]{4}', String.Escape),
            # Yes, \U literals are 6 hex digits.
            (r'\\U[0-9a-fA-F]{6}', String.Escape)
        ]
    }



# -*- coding: utf-8 -*-
"""
    pygments.lexers._stata_builtins
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Builtins for Stata

    :copyright: Copyright 2006-2017 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""


builtins_base = (
    "if", "else", "in", "foreach", "for", "forv", "forva",
    "forval", "forvalu", "forvalue", "forvalues", "by", "bys",
    "bysort", "quietly", "qui", "about", "ac",
    "ac_7", "acprplot", "acprplot_7", "adjust", "ado", "adopath",
    "adoupdate", "alpha", "ameans", "an", "ano", "anov", "anova",
    "anova_estat", "anova_terms", "anovadef", "aorder", "ap", "app",
    "appe", "appen", "append", "arch", "arch_dr", "arch_estat",
    "arch_p", "archlm", "areg", "areg_p", "args", "arima",
    "arima_dr", "arima_estat", "arima_p", "as", "asmprobit",
    "asmprobit_estat", "asmprobit_lf", "asmprobit_mfx__dlg",
    "asmprobit_p", "ass", "asse", "asser", "assert", "avplot",
    "avplot_7", "avplots", "avplots_7", "bcskew0", "bgodfrey",
    "binreg", "bip0_lf", "biplot", "bipp_lf", "bipr_lf",
    "bipr_p", "biprobit", "bitest", "bitesti", "bitowt", "blogit",
    "bmemsize", "boot", "bootsamp", "bootstrap", "bootstrap_8",
    "boxco_l", "boxco_p", "boxcox", "boxcox_6", "boxcox_p",
    "bprobit", "br", "break", "brier", "bro", "brow", "brows",
    "browse", "brr", "brrstat", "bs", "bs_7", "bsampl_w",
    "bsample", "bsample_7", "bsqreg", "bstat", "bstat_7", "bstat_8",
    "bstrap", "bstrap_7", "ca", "ca_estat", "ca_p", "cabiplot",
    "camat", "canon", "canon_8", "canon_8_p", "canon_estat",
    "canon_p", "cap", "caprojection", "capt", "captu", "captur",
    "capture", "cat", "cc", "cchart", "cchart_7", "cci",
    "cd", "censobs_table", "centile", "cf", "char", "chdir",
    "checkdlgfiles", "checkestimationsample", "checkhlpfiles",
    "checksum", "chelp", "ci", "cii", "cl", "class", "classutil",
    "clear", "cli", "clis", "clist", "clo", "clog", "clog_lf",
    "clog_p", "clogi", "clogi_sw", "clogit", "clogit_lf",
    "clogit_p", "clogitp", "clogl_sw", "cloglog", "clonevar",
    "clslistarray", "cluster", "cluster_measures", "cluster_stop",
    "cluster_tree", "cluster_tree_8", "clustermat", "cmdlog",
    "cnr", "cnre", "cnreg", "cnreg_p", "cnreg_sw", "cnsreg",
    "codebook", "collaps4", "collapse", "colormult_nb",
    "colormult_nw", "compare", "compress", "conf", "confi",
    "confir", "confirm", "conren", "cons", "const", "constr",
    "constra", "constrai", "constrain", "constraint", "continue",
    "contract", "copy", "copyright", "copysource", "cor", "corc",
    "corr", "corr2data", "corr_anti", "corr_kmo", "corr_smc",
    "corre", "correl", "correla", "correlat", "correlate",
    "corrgram", "cou", "coun", "count", "cox", "cox_p", "cox_sw",
    "coxbase", "coxhaz", "coxvar", "cprplot", "cprplot_7",
    "crc", "cret", "cretu", "cretur", "creturn", "cross", "cs",
    "cscript", "cscript_log", "csi", "ct", "ct_is", "ctset",
    "ctst_5", "ctst_st", "cttost", "cumsp", "cumsp_7", "cumul",
    "cusum", "cusum_7", "cutil", "d", "datasig", "datasign",
    "datasigna", "datasignat", "datasignatu", "datasignatur",
    "datasignature", "datetof", "db", "dbeta", "de", "dec",
    "deco", "decod", "decode", "deff", "des", "desc", "descr",
    "descri", "describ", "describe", "destring", "dfbeta",
    "dfgls", "dfuller", "di", "di_g", "dir", "dirstats", "dis",
    "discard", "disp", "disp_res", "disp_s", "displ", "displa",
    "display", "distinct", "do", "doe", "doed", "doedi",
    "doedit", "dotplot", "dotplot_7", "dprobit", "drawnorm",
    "drop", "ds", "ds_util", "dstdize", "duplicates", "durbina",
    "dwstat", "dydx", "e", "ed", "edi", "edit", "egen",
    "eivreg", "emdef", "en", "enc", "enco", "encod", "encode",
    "eq", "erase", "ereg", "ereg_lf", "ereg_p", "ereg_sw",
    "ereghet", "ereghet_glf", "ereghet_glf_sh", "ereghet_gp",
    "ereghet_ilf", "ereghet_ilf_sh", "ereghet_ip", "eret",
    "eretu", "eretur", "ereturn", "err", "erro", "error", "est",
    "est_cfexist", "est_cfname", "est_clickable", "est_expand",
    "est_hold", "est_table", "est_unhold", "est_unholdok",
    "estat", "estat_default", "estat_summ", "estat_vce_only",
    "esti", "estimates", "etodow", "etof", "etomdy", "ex",
    "exi", "exit", "expand", "expandcl", "fac", "fact", "facto",
    "factor", "factor_estat", "factor_p", "factor_pca_rotated",
    "factor_rotate", "factormat", "fcast", "fcast_compute",
    "fcast_graph", "fdades", "fdadesc", "fdadescr", "fdadescri",
    "fdadescrib", "fdadescribe", "fdasav", "fdasave", "fdause",
    "fh_st", "open", "read", "close",
    "file", "filefilter", "fillin", "find_hlp_file", "findfile",
    "findit", "findit_7", "fit", "fl", "fli", "flis", "flist",
    "for5_0", "form", "forma", "format", "fpredict", "frac_154",
    "frac_adj", "frac_chk", "frac_cox", "frac_ddp", "frac_dis",
    "frac_dv", "frac_in", "frac_mun", "frac_pp", "frac_pq",
    "frac_pv", "frac_wgt", "frac_xo", "fracgen", "fracplot",
    "fracplot_7", "fracpoly", "fracpred", "fron_ex", "fron_hn",
    "fron_p", "fron_tn", "fron_tn2", "frontier", "ftodate", "ftoe",
    "ftomdy", "ftowdate", "g", "gamhet_glf", "gamhet_gp",
    "gamhet_ilf", "gamhet_ip", "gamma", "gamma_d2", "gamma_p",
    "gamma_sw", "gammahet", "gdi_hexagon", "gdi_spokes", "ge",
    "gen", "gene", "gener", "genera", "generat", "generate",
    "genrank", "genstd", "genvmean", "gettoken", "gl", "gladder",
    "gladder_7", "glim_l01", "glim_l02", "glim_l03", "glim_l04",
    "glim_l05", "glim_l06", "glim_l07", "glim_l08", "glim_l09",
    "glim_l10", "glim_l11", "glim_l12", "glim_lf", "glim_mu",
    "glim_nw1", "glim_nw2", "glim_nw3", "glim_p", "glim_v1",
    "glim_v2", "glim_v3", "glim_v4", "glim_v5", "glim_v6",
    "glim_v7", "glm", "glm_6", "glm_p", "glm_sw", "glmpred", "glo",
    "glob", "globa", "global", "glogit", "glogit_8", "glogit_p",
    "gmeans", "gnbre_lf", "gnbreg", "gnbreg_5", "gnbreg_p",
    "gomp_lf", "gompe_sw", "gomper_p", "gompertz", "gompertzhet",
    "gomphet_glf", "gomphet_glf_sh", "gomphet_gp", "gomphet_ilf",
    "gomphet_ilf_sh", "gomphet_ip", "gphdot", "gphpen",
    "gphprint", "gprefs", "gprobi_p", "gprobit", "gprobit_8", "gr",
    "gr7", "gr_copy", "gr_current", "gr_db", "gr_describe",
    "gr_dir", "gr_draw", "gr_draw_replay", "gr_drop", "gr_edit",
    "gr_editviewopts", "gr_example", "gr_example2", "gr_export",
    "gr_print", "gr_qscheme", "gr_query", "gr_read", "gr_rename",
    "gr_replay", "gr_save", "gr_set", "gr_setscheme", "gr_table",
    "gr_undo", "gr_use", "graph", "graph7", "grebar", "greigen",
    "greigen_7", "greigen_8", "grmeanby", "grmeanby_7",
    "gs_fileinfo", "gs_filetype", "gs_graphinfo", "gs_stat",
    "gsort", "gwood", "h", "hadimvo", "hareg", "hausman",
    "haver", "he", "heck_d2", "heckma_p", "heckman", "heckp_lf",
    "heckpr_p", "heckprob", "hel", "help", "hereg", "hetpr_lf",
    "hetpr_p", "hetprob", "hettest", "hexdump", "hilite",
    "hist", "hist_7", "histogram", "hlogit", "hlu", "hmeans",
    "hotel", "hotelling", "hprobit", "hreg", "hsearch", "icd9",
    "icd9_ff", "icd9p", "iis", "impute", "imtest", "inbase",
    "include", "inf", "infi", "infil", "infile", "infix", "inp",
    "inpu", "input", "ins", "insheet", "insp", "inspe",
    "inspec", "inspect", "integ", "inten", "intreg", "intreg_7",
    "intreg_p", "intrg2_ll", "intrg_ll", "intrg_ll2", "ipolate",
    "iqreg", "ir", "irf", "irf_create", "irfm", "iri", "is_svy",
    "is_svysum", "isid", "istdize", "ivprob_1_lf", "ivprob_lf",
    "ivprobit", "ivprobit_p", "ivreg", "ivreg_footnote",
    "ivtob_1_lf", "ivtob_lf", "ivtobit", "ivtobit_p", "jackknife",
    "jacknife", "jknife", "jknife_6", "jknife_8", "jkstat",
    "joinby", "kalarma1", "kap", "kap_3", "kapmeier", "kappa",
    "kapwgt", "kdensity", "kdensity_7", "keep", "ksm", "ksmirnov",
    "ktau", "kwallis", "l", "la", "lab", "labe", "label",
    "labelbook", "ladder", "levels", "levelsof", "leverage",
    "lfit", "lfit_p", "li", "lincom", "line", "linktest",
    "lis", "list", "lloghet_glf", "lloghet_glf_sh", "lloghet_gp",
    "lloghet_ilf", "lloghet_ilf_sh", "lloghet_ip", "llogi_sw",
    "llogis_p", "llogist", "llogistic", "llogistichet",
    "lnorm_lf", "lnorm_sw", "lnorma_p", "lnormal", "lnormalhet",
    "lnormhet_glf", "lnormhet_glf_sh", "lnormhet_gp",
    "lnormhet_ilf", "lnormhet_ilf_sh", "lnormhet_ip", "lnskew0",
    "loadingplot", "loc", "loca", "local", "log", "logi",
    "logis_lf", "logistic", "logistic_p", "logit", "logit_estat",
    "logit_p", "loglogs", "logrank", "loneway", "lookfor",
    "lookup", "lowess", "lowess_7", "lpredict", "lrecomp", "lroc",
    "lroc_7", "lrtest", "ls", "lsens", "lsens_7", "lsens_x",
    "lstat", "ltable", "ltable_7", "ltriang", "lv", "lvr2plot",
    "lvr2plot_7", "m", "ma", "mac", "macr", "macro", "makecns",
    "man", "manova", "manova_estat", "manova_p", "manovatest",
    "mantel", "mark", "markin", "markout", "marksample", "mat",
    "mat_capp", "mat_order", "mat_put_rr", "mat_rapp", "mata",
    "mata_clear", "mata_describe", "mata_drop", "mata_matdescribe",
    "mata_matsave", "mata_matuse", "mata_memory", "mata_mlib",
    "mata_mosave", "mata_rename", "mata_which", "matalabel",
    "matcproc", "matlist", "matname", "matr", "matri",
    "matrix", "matrix_input__dlg", "matstrik", "mcc", "mcci",
    "md0_", "md1_", "md1debug_", "md2_", "md2debug_", "mds",
    "mds_estat", "mds_p", "mdsconfig", "mdslong", "mdsmat",
    "mdsshepard", "mdytoe", "mdytof", "me_derd", "mean",
    "means", "median", "memory", "memsize", "meqparse", "mer",
    "merg", "merge", "mfp", "mfx", "mhelp", "mhodds", "minbound",
    "mixed_ll", "mixed_ll_reparm", "mkassert", "mkdir",
    "mkmat", "mkspline", "ml", "ml_5", "ml_adjs", "ml_bhhhs",
    "ml_c_d", "ml_check", "ml_clear", "ml_cnt", "ml_debug",
    "ml_defd", "ml_e0", "ml_e0_bfgs", "ml_e0_cycle", "ml_e0_dfp",
    "ml_e0i", "ml_e1", "ml_e1_bfgs", "ml_e1_bhhh", "ml_e1_cycle",
    "ml_e1_dfp", "ml_e2", "ml_e2_cycle", "ml_ebfg0", "ml_ebfr0",
    "ml_ebfr1", "ml_ebh0q", "ml_ebhh0", "ml_ebhr0", "ml_ebr0i",
    "ml_ecr0i", "ml_edfp0", "ml_edfr0", "ml_edfr1", "ml_edr0i",
    "ml_eds", "ml_eer0i", "ml_egr0i", "ml_elf", "ml_elf_bfgs",
    "ml_elf_bhhh", "ml_elf_cycle", "ml_elf_dfp", "ml_elfi",
    "ml_elfs", "ml_enr0i", "ml_enrr0", "ml_erdu0", "ml_erdu0_bfgs",
    "ml_erdu0_bhhh", "ml_erdu0_bhhhq", "ml_erdu0_cycle",
    "ml_erdu0_dfp", "ml_erdu0_nrbfgs", "ml_exde", "ml_footnote",
    "ml_geqnr", "ml_grad0", "ml_graph", "ml_hbhhh", "ml_hd0",
    "ml_hold", "ml_init", "ml_inv", "ml_log", "ml_max",
    "ml_mlout", "ml_mlout_8", "ml_model", "ml_nb0", "ml_opt",
    "ml_p", "ml_plot", "ml_query", "ml_rdgrd", "ml_repor",
    "ml_s_e", "ml_score", "ml_searc", "ml_technique", "ml_unhold",
    "mleval", "mlf_", "mlmatbysum", "mlmatsum", "mlog", "mlogi",
    "mlogit", "mlogit_footnote", "mlogit_p", "mlopts", "mlsum",
    "mlvecsum", "mnl0_", "mor", "more", "mov", "move", "mprobit",
    "mprobit_lf", "mprobit_p", "mrdu0_", "mrdu1_", "mvdecode",
    "mvencode", "mvreg", "mvreg_estat", "n", "nbreg",
    "nbreg_al", "nbreg_lf", "nbreg_p", "nbreg_sw", "nestreg", "net",
    "newey", "newey_7", "newey_p", "news", "nl", "nl_7", "nl_9",
    "nl_9_p", "nl_p", "nl_p_7", "nlcom", "nlcom_p", "nlexp2",
    "nlexp2_7", "nlexp2a", "nlexp2a_7", "nlexp3", "nlexp3_7",
    "nlgom3", "nlgom3_7", "nlgom4", "nlgom4_7", "nlinit", "nllog3",
    "nllog3_7", "nllog4", "nllog4_7", "nlog_rd", "nlogit",
    "nlogit_p", "nlogitgen", "nlogittree", "nlpred", "no",
    "nobreak", "noi", "nois", "noisi", "noisil", "noisily", "note",
    "notes", "notes_dlg", "nptrend", "numlabel", "numlist", "odbc",
    "old_ver", "olo", "olog", "ologi", "ologi_sw", "ologit",
    "ologit_p", "ologitp", "on", "one", "onew", "onewa", "oneway",
    "op_colnm", "op_comp", "op_diff", "op_inv", "op_str", "opr",
    "opro", "oprob", "oprob_sw", "oprobi", "oprobi_p", "oprobit",
    "oprobitp", "opts_exclusive", "order", "orthog", "orthpoly",
    "ou", "out", "outf", "outfi", "outfil", "outfile", "outs",
    "outsh", "outshe", "outshee", "outsheet", "ovtest", "pac",
    "pac_7", "palette", "parse", "parse_dissim", "pause", "pca",
    "pca_8", "pca_display", "pca_estat", "pca_p", "pca_rotate",
    "pcamat", "pchart", "pchart_7", "pchi", "pchi_7", "pcorr",
    "pctile", "pentium", "pergram", "pergram_7", "permute",
    "permute_8", "personal", "peto_st", "pkcollapse", "pkcross",
    "pkequiv", "pkexamine", "pkexamine_7", "pkshape", "pksumm",
    "pksumm_7", "pl", "plo", "plot", "plugin", "pnorm",
    "pnorm_7", "poisgof", "poiss_lf", "poiss_sw", "poisso_p",
    "poisson", "poisson_estat", "post", "postclose", "postfile",
    "postutil", "pperron", "pr", "prais", "prais_e", "prais_e2",
    "prais_p", "predict", "predictnl", "preserve", "print",
    "pro", "prob", "probi", "probit", "probit_estat", "probit_p",
    "proc_time", "procoverlay", "procrustes", "procrustes_estat",
    "procrustes_p", "profiler", "prog", "progr", "progra",
    "program", "prop", "proportion", "prtest", "prtesti", "pwcorr",
    "pwd", "q", "s", "qby", "qbys", "qchi", "qchi_7", "qladder",
    "qladder_7", "qnorm", "qnorm_7", "qqplot", "qqplot_7", "qreg",
    "qreg_c", "qreg_p", "qreg_sw", "qu", "quadchk", "quantile",
    "quantile_7", "que", "quer", "query", "range", "ranksum",
    "ratio", "rchart", "rchart_7", "rcof", "recast", "reclink",
    "recode", "reg", "reg3", "reg3_p", "regdw", "regr", "regre",
    "regre_p2", "regres", "regres_p", "regress", "regress_estat",
    "regriv_p", "remap", "ren", "rena", "renam", "rename",
    "renpfix", "repeat", "replace", "report", "reshape",
    "restore", "ret", "retu", "retur", "return", "rm", "rmdir",
    "robvar", "roccomp", "roccomp_7", "roccomp_8", "rocf_lf",
    "rocfit", "rocfit_8", "rocgold", "rocplot", "rocplot_7",
    "roctab", "roctab_7", "rolling", "rologit", "rologit_p",
    "rot", "rota", "rotat", "rotate", "rotatemat", "rreg",
    "rreg_p", "ru", "run", "runtest", "rvfplot", "rvfplot_7",
    "rvpplot", "rvpplot_7", "sa", "safesum", "sample",
    "sampsi", "sav", "save", "savedresults", "saveold", "sc",
    "sca", "scal", "scala", "scalar", "scatter", "scm_mine",
    "sco", "scob_lf", "scob_p", "scobi_sw", "scobit", "scor",
    "score", "scoreplot", "scoreplot_help", "scree", "screeplot",
    "screeplot_help", "sdtest", "sdtesti", "se", "search",
    "separate", "seperate", "serrbar", "serrbar_7", "serset", "set",
    "set_defaults", "sfrancia", "sh", "she", "shel", "shell",
    "shewhart", "shewhart_7", "signestimationsample", "signrank",
    "signtest", "simul", "simul_7", "simulate", "simulate_8",
    "sktest", "sleep", "slogit", "slogit_d2", "slogit_p", "smooth",
    "snapspan", "so", "sor", "sort", "spearman", "spikeplot",
    "spikeplot_7", "spikeplt", "spline_x", "split", "sqreg",
    "sqreg_p", "sret", "sretu", "sretur", "sreturn", "ssc", "st",
    "st_ct", "st_hc", "st_hcd", "st_hcd_sh", "st_is", "st_issys",
    "st_note", "st_promo", "st_set", "st_show", "st_smpl",
    "st_subid", "stack", "statsby", "statsby_8", "stbase", "stci",
    "stci_7", "stcox", "stcox_estat", "stcox_fr", "stcox_fr_ll",
    "stcox_p", "stcox_sw", "stcoxkm", "stcoxkm_7", "stcstat",
    "stcurv", "stcurve", "stcurve_7", "stdes", "stem", "stepwise",
    "stereg", "stfill", "stgen", "stir", "stjoin", "stmc", "stmh",
    "stphplot", "stphplot_7", "stphtest", "stphtest_7",
    "stptime", "strate", "strate_7", "streg", "streg_sw", "streset",
    "sts", "sts_7", "stset", "stsplit", "stsum", "sttocc",
    "sttoct", "stvary", "stweib", "su", "suest", "suest_8",
    "sum", "summ", "summa", "summar", "summari", "summariz",
    "summarize", "sunflower", "sureg", "survcurv", "survsum",
    "svar", "svar_p", "svmat", "svy", "svy_disp", "svy_dreg",
    "svy_est", "svy_est_7", "svy_estat", "svy_get", "svy_gnbreg_p",
    "svy_head", "svy_header", "svy_heckman_p", "svy_heckprob_p",
    "svy_intreg_p", "svy_ivreg_p", "svy_logistic_p", "svy_logit_p",
    "svy_mlogit_p", "svy_nbreg_p", "svy_ologit_p", "svy_oprobit_p",
    "svy_poisson_p", "svy_probit_p", "svy_regress_p", "svy_sub",
    "svy_sub_7", "svy_x", "svy_x_7", "svy_x_p", "svydes",
    "svydes_8", "svygen", "svygnbreg", "svyheckman", "svyheckprob",
    "svyintreg", "svyintreg_7", "svyintrg", "svyivreg", "svylc",
    "svylog_p", "svylogit", "svymarkout", "svymarkout_8",
    "svymean", "svymlog", "svymlogit", "svynbreg", "svyolog",
    "svyologit", "svyoprob", "svyoprobit", "svyopts",
    "svypois", "svypois_7", "svypoisson", "svyprobit", "svyprobt",
    "svyprop", "svyprop_7", "svyratio", "svyreg", "svyreg_p",
    "svyregress", "svyset", "svyset_7", "svyset_8", "svytab",
    "svytab_7", "svytest", "svytotal", "sw", "sw_8", "swcnreg",
    "swcox", "swereg", "swilk", "swlogis", "swlogit",
    "swologit", "swoprbt", "swpois", "swprobit", "swqreg",
    "swtobit", "swweib", "symmetry", "symmi", "symplot",
    "symplot_7", "syntax", "sysdescribe", "sysdir", "sysuse",
    "szroeter", "ta", "tab", "tab1", "tab2", "tab_or", "tabd",
    "tabdi", "tabdis", "tabdisp", "tabi", "table", "tabodds",
    "tabodds_7", "tabstat", "tabu", "tabul", "tabula", "tabulat",
    "tabulate", "te", "tempfile", "tempname", "tempvar", "tes",
    "test", "testnl", "testparm", "teststd", "tetrachoric",
    "time_it", "timer", "tis", "tob", "tobi", "tobit", "tobit_p",
    "tobit_sw", "token", "tokeni", "tokeniz", "tokenize",
    "tostring", "total", "translate", "translator", "transmap",
    "treat_ll", "treatr_p", "treatreg", "trim", "trnb_cons",
    "trnb_mean", "trpoiss_d2", "trunc_ll", "truncr_p", "truncreg",
    "tsappend", "tset", "tsfill", "tsline", "tsline_ex",
    "tsreport", "tsrevar", "tsrline", "tsset", "tssmooth",
    "tsunab", "ttest", "ttesti", "tut_chk", "tut_wait", "tutorial",
    "tw", "tware_st", "two", "twoway", "twoway__fpfit_serset",
    "twoway__function_gen", "twoway__histogram_gen",
    "twoway__ipoint_serset", "twoway__ipoints_serset",
    "twoway__kdensity_gen", "twoway__lfit_serset",
    "twoway__normgen_gen", "twoway__pci_serset",
    "twoway__qfit_serset", "twoway__scatteri_serset",
    "twoway__sunflower_gen", "twoway_ksm_serset", "ty", "typ",
    "type", "typeof", "u", "unab", "unabbrev", "unabcmd",
    "update", "us", "use", "uselabel", "var", "var_mkcompanion",
    "var_p", "varbasic", "varfcast", "vargranger", "varirf",
    "varirf_add", "varirf_cgraph", "varirf_create", "varirf_ctable",
    "varirf_describe", "varirf_dir", "varirf_drop", "varirf_erase",
    "varirf_graph", "varirf_ograph", "varirf_rename", "varirf_set",
    "varirf_table", "varlist", "varlmar", "varnorm", "varsoc",
    "varstable", "varstable_w", "varstable_w2", "varwle",
    "vce", "vec", "vec_fevd", "vec_mkphi", "vec_p", "vec_p_w",
    "vecirf_create", "veclmar", "veclmar_w", "vecnorm",
    "vecnorm_w", "vecrank", "vecstable", "verinst", "vers",
    "versi", "versio", "version", "view", "viewsource", "vif",
    "vwls", "wdatetof", "webdescribe", "webseek", "webuse",
    "weib1_lf", "weib2_lf", "weib_lf", "weib_lf0", "weibhet_glf",
    "weibhet_glf_sh", "weibhet_glfa", "weibhet_glfa_sh",
    "weibhet_gp", "weibhet_ilf", "weibhet_ilf_sh", "weibhet_ilfa",
    "weibhet_ilfa_sh", "weibhet_ip", "weibu_sw", "weibul_p",
    "weibull", "weibull_c", "weibull_s", "weibullhet",
    "wh", "whelp", "whi", "which", "whil", "while", "wilc_st",
    "wilcoxon", "win", "wind", "windo", "window", "winexec",
    "wntestb", "wntestb_7", "wntestq", "xchart", "xchart_7",
    "xcorr", "xcorr_7", "xi", "xi_6", "xmlsav", "xmlsave",
    "xmluse", "xpose", "xsh", "xshe", "xshel", "xshell",
    "xt_iis", "xt_tis", "xtab_p", "xtabond", "xtbin_p",
    "xtclog", "xtcloglog", "xtcloglog_8", "xtcloglog_d2",
    "xtcloglog_pa_p", "xtcloglog_re_p", "xtcnt_p", "xtcorr",
    "xtdata", "xtdes", "xtfront_p", "xtfrontier", "xtgee",
    "xtgee_elink", "xtgee_estat", "xtgee_makeivar", "xtgee_p",
    "xtgee_plink", "xtgls", "xtgls_p", "xthaus", "xthausman",
    "xtht_p", "xthtaylor", "xtile", "xtint_p", "xtintreg",
    "xtintreg_8", "xtintreg_d2", "xtintreg_p", "xtivp_1",
    "xtivp_2", "xtivreg", "xtline", "xtline_ex", "xtlogit",
    "xtlogit_8", "xtlogit_d2", "xtlogit_fe_p", "xtlogit_pa_p",
    "xtlogit_re_p", "xtmixed", "xtmixed_estat", "xtmixed_p",
    "xtnb_fe", "xtnb_lf", "xtnbreg", "xtnbreg_pa_p",
    "xtnbreg_refe_p", "xtpcse", "xtpcse_p", "xtpois", "xtpoisson",
    "xtpoisson_d2", "xtpoisson_pa_p", "xtpoisson_refe_p", "xtpred",
    "xtprobit", "xtprobit_8", "xtprobit_d2", "xtprobit_re_p",
    "xtps_fe", "xtps_lf", "xtps_ren", "xtps_ren_8", "xtrar_p",
    "xtrc", "xtrc_p", "xtrchh", "xtrefe_p", "xtreg", "xtreg_be",
    "xtreg_fe", "xtreg_ml", "xtreg_pa_p", "xtreg_re",
    "xtregar", "xtrere_p", "xtset", "xtsf_ll", "xtsf_llti",
    "xtsum", "xttab", "xttest0", "xttobit", "xttobit_8",
    "xttobit_p", "xttrans", "yx", "yxview__barlike_draw",
    "yxview_area_draw", "yxview_bar_draw", "yxview_dot_draw",
    "yxview_dropline_draw", "yxview_function_draw",
    "yxview_iarrow_draw", "yxview_ilabels_draw",
    "yxview_normal_draw", "yxview_pcarrow_draw",
    "yxview_pcbarrow_draw", "yxview_pccapsym_draw",
    "yxview_pcscatter_draw", "yxview_pcspike_draw",
    "yxview_rarea_draw", "yxview_rbar_draw", "yxview_rbarm_draw",
    "yxview_rcap_draw", "yxview_rcapsym_draw",
    "yxview_rconnected_draw", "yxview_rline_draw",
    "yxview_rscatter_draw", "yxview_rspike_draw",
    "yxview_spike_draw", "yxview_sunflower_draw", "zap_s", "zinb",
    "zinb_llf", "zinb_plf", "zip", "zip_llf", "zip_p", "zip_plf",
    "zt_ct_5", "zt_hc_5", "zt_hcd_5", "zt_is_5", "zt_iss_5",
    "zt_sho_5", "zt_smp_5", "ztbase_5", "ztcox_5", "ztdes_5",
    "ztereg_5", "ztfill_5", "ztgen_5", "ztir_5", "ztjoin_5", "ztnb",
    "ztnb_p", "ztp", "ztp_p", "zts_5", "ztset_5", "ztspli_5",
    "ztsum_5", "zttoct_5", "ztvary_5", "ztweib_5"
)

builtins_functions = (
    "Cdhms", "Chms", "Clock", "Cmdyhms", "Cofc", "Cofd", "F",
    "Fden", "Ftail", "I", "J", "_caller", "abbrev", "abs", "acos",
    "acosh", "asin", "asinh", "atan", "atan2", "atanh",
    "autocode", "betaden", "binomial", "binomialp", "binomialtail",
    "binormal", "bofd", "byteorder", "c", "ceil", "char",
    "chi2", "chi2den", "chi2tail", "cholesky", "chop", "clip",
    "clock", "cloglog", "cofC", "cofd", "colnumb", "colsof", "comb",
    "cond", "corr", "cos", "cosh", "d", "daily", "date", "day",
    "det", "dgammapda", "dgammapdada", "dgammapdadx", "dgammapdx",
    "dgammapdxdx", "dhms", "diag", "diag0cnt", "digamma",
    "dofC", "dofb", "dofc", "dofh", "dofm", "dofq", "dofw",
    "dofy", "dow", "doy", "dunnettprob", "e", "el", "epsdouble",
    "epsfloat", "exp", "fileexists", "fileread", "filereaderror",
    "filewrite", "float", "floor", "fmtwidth", "gammaden",
    "gammap", "gammaptail", "get", "group", "h", "hadamard",
    "halfyear", "halfyearly", "has_eprop", "hh", "hhC", "hms",
    "hofd", "hours", "hypergeometric", "hypergeometricp", "ibeta",
    "ibetatail", "index", "indexnot", "inlist", "inrange", "int",
    "inv", "invF", "invFtail", "invbinomial", "invbinomialtail",
    "invchi2", "invchi2tail", "invcloglog", "invdunnettprob",
    "invgammap", "invgammaptail", "invibeta", "invibetatail",
    "invlogit", "invnFtail", "invnbinomial", "invnbinomialtail",
    "invnchi2", "invnchi2tail", "invnibeta", "invnorm", "invnormal",
    "invnttail", "invpoisson", "invpoissontail", "invsym", "invt",
    "invttail", "invtukeyprob", "irecode", "issym", "issymmetric",
    "itrim", "length", "ln", "lnfact", "lnfactorial", "lngamma",
    "lnnormal", "lnnormalden", "log", "log10", "logit", "lower",
    "ltrim", "m", "match", "matmissing", "matrix", "matuniform",
    "max", "maxbyte", "maxdouble", "maxfloat", "maxint", "maxlong",
    "mdy", "mdyhms", "mi", "min", "minbyte", "mindouble",
    "minfloat", "minint", "minlong", "minutes", "missing", "mm",
    "mmC", "mod", "mofd", "month", "monthly", "mreldif",
    "msofhours", "msofminutes", "msofseconds", "nF", "nFden",
    "nFtail", "nbetaden", "nbinomial", "nbinomialp", "nbinomialtail",
    "nchi2", "nchi2den", "nchi2tail", "nibeta", "norm", "normal",
    "normalden", "normd", "npnF", "npnchi2", "npnt", "nt", "ntden",
    "nttail", "nullmat", "plural", "poisson", "poissonp",
    "poissontail", "proper", "q", "qofd", "quarter", "quarterly",
    "r", "rbeta", "rbinomial", "rchi2", "real", "recode", "regexm",
    "regexr", "regexs", "reldif", "replay", "return", "reverse",
    "rgamma", "rhypergeometric", "rnbinomial", "rnormal", "round",
    "rownumb", "rowsof", "rpoisson", "rt", "rtrim", "runiform", "s",
    "scalar", "seconds", "sign", "sin", "sinh", "smallestdouble",
    "soundex", "soundex_nara", "sqrt", "ss", "ssC", "strcat",
    "strdup", "string", "strlen", "strlower", "strltrim", "strmatch",
    "strofreal", "strpos", "strproper", "strreverse", "strrtrim",
    "strtoname", "strtrim", "strupper", "subinstr", "subinword",
    "substr", "sum", "sweep", "syminv", "t", "tC", "tan", "tanh",
    "tc", "td", "tden", "th", "tin", "tm", "tq", "trace",
    "trigamma", "trim", "trunc", "ttail", "tukeyprob", "tw",
    "twithin", "uniform", "upper", "vec", "vecdiag", "w", "week",
    "weekly", "wofd", "word", "wordcount", "year", "yearly",
    "yh", "ym", "yofd", "yq", "yw"
)





# -*- coding: utf-8 -*-
"""
    wakatime.dependencies
    ~~~~~~~~~~~~~~~~~~~~~

    Parse dependencies from a source code file.

    :copyright: (c) 2014 Alan Hamlett.
    :license: BSD, see LICENSE for more details.
"""

import logging
import re
import sys

from ..compat import u, open, import_module
from ..exceptions import NotYetImplemented


log = logging.getLogger('WakaTime')


class TokenParser(object):
    """The base class for all dependency parsers. To add support for your
    language, inherit from this class and implement the :meth:`parse` method
    to return a list of dependency strings.
    """
    exclude = []

    def __init__(self, source_file, lexer=None):
        self._tokens = None
        self.dependencies = []
        self.source_file = source_file
        self.lexer = lexer
        self.exclude = [re.compile(x, re.IGNORECASE) for x in self.exclude]

    @property
    def tokens(self):
        if self._tokens is None:
            self._tokens = self._extract_tokens()
        return self._tokens

    def parse(self, tokens=[]):
        """ Should return a list of dependencies.
        """
        raise NotYetImplemented()

    def append(self, dep, truncate=False, separator=None, truncate_to=None,
               strip_whitespace=True):
        self._save_dependency(
            dep,
            truncate=truncate,
            truncate_to=truncate_to,
            separator=separator,
            strip_whitespace=strip_whitespace,
        )

    def partial(self, token):
        return u(token).split('.')[-1]

    def _extract_tokens(self):
        if self.lexer:
            try:
                with open(self.source_file, 'r', encoding='utf-8') as fh:
                    return self.lexer.get_tokens_unprocessed(fh.read(512000))
            except:
                pass
            try:
                with open(self.source_file, 'r', encoding=sys.getfilesystemencoding()) as fh:
                    return self.lexer.get_tokens_unprocessed(fh.read(512000))  # pragma: nocover
            except:
                pass
        return []

    def _save_dependency(self, dep, truncate=False, separator=None,
                         truncate_to=None, strip_whitespace=True):
        if truncate:
            if separator is None:
                separator = u('.')
            separator = u(separator)
            dep = dep.split(separator)
            if truncate_to is None or truncate_to < 1:
                truncate_to = 1
            if truncate_to > len(dep):
                truncate_to = len(dep)
            dep = dep[0] if len(dep) == 1 else separator.join(dep[:truncate_to])
        if strip_whitespace:
            dep = dep.strip()
        if dep and (not separator or not dep.startswith(separator)):
            should_exclude = False
            for compiled in self.exclude:
                if compiled.search(dep):
                    should_exclude = True
                    break
            if not should_exclude:
                self.dependencies.append(dep)


class DependencyParser(object):
    source_file = None
    lexer = None
    parser = None

    def __init__(self, source_file, lexer):
        self.source_file = source_file
        self.lexer = lexer

        if self.lexer:
            module_name = self.lexer.__module__.rsplit('.', 1)[-1]
            class_name = self.lexer.__class__.__name__.replace('Lexer', 'Parser', 1)
        else:
            module_name = 'unknown'
            class_name = 'UnknownParser'

        try:
            module = import_module('.%s' % module_name, package=__package__)
            try:
                self.parser = getattr(module, class_name)
            except AttributeError:
                log.debug('Parsing dependencies not supported for {0}.{1}'.format(module_name, class_name))
        except ImportError:
            log.debug('Parsing dependencies not supported for {0}.{1}'.format(module_name, class_name))

    def parse(self):
        if self.parser:
            plugin = self.parser(self.source_file, lexer=self.lexer)
            dependencies = plugin.parse()
            return list(set(dependencies))
        return []



# -*- coding: utf-8 -*-
"""
    wakatime.languages.dotnet
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Parse dependencies from .NET code.

    :copyright: (c) 2014 Alan Hamlett.
    :license: BSD, see LICENSE for more details.
"""

from . import TokenParser
from ..compat import u


class CSharpParser(TokenParser):
    exclude = [
        r'^system$',
        r'^microsoft$',
    ]
    state = None
    buffer = u('')

    def parse(self):
        for index, token, content in self.tokens:
            self._process_token(token, content)
        return self.dependencies

    def _process_token(self, token, content):
        if self.partial(token) == 'Keyword':
            self._process_keyword(token, content)
        if self.partial(token) == 'Namespace' or self.partial(token) == 'Name':
            self._process_namespace(token, content)
        elif self.partial(token) == 'Punctuation':
            self._process_punctuation(token, content)
        else:
            self._process_other(token, content)

    def _process_keyword(self, token, content):
        if content == 'using':
            self.state = 'import'
            self.buffer = u('')

    def _process_namespace(self, token, content):
        if self.state == 'import':
            if u(content) != u('import') and u(content) != u('package') and u(content) != u('namespace') and u(content) != u('static'):
                if u(content) == u(';'):  # pragma: nocover
                    self._process_punctuation(token, content)
                else:
                    self.buffer += u(content)

    def _process_punctuation(self, token, content):
        if self.state == 'import':
            if u(content) == u(';'):
                self.append(self.buffer, truncate=True)
                self.buffer = u('')
                self.state = None
            elif u(content) == u('='):
                self.buffer = u('')
            else:
                self.buffer += u(content)

    def _process_other(self, token, content):
        pass



# -*- coding: utf-8 -*-
"""
    wakatime.languages.unknown
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    Parse dependencies from files of unknown language.

    :copyright: (c) 2014 Alan Hamlett.
    :license: BSD, see LICENSE for more details.
"""

import os

from . import TokenParser


FILES = {
    'bower': {'exact': False, 'dependency': 'bower'},
    'grunt': {'exact': False, 'dependency': 'grunt'},
}


class UnknownParser(TokenParser):

    def parse(self):
        self._process_file_name(os.path.basename(self.source_file))
        return self.dependencies

    def _process_file_name(self, file_name):
        for key, value in FILES.items():
            found = (key == file_name) if value.get('exact') else (key.lower() in file_name.lower())
            if found:
                self.append(value['dependency'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
certifi.py
~~~~~~~~~~

This module returns the installation location of cacert.pem.
"""
import os
import warnings


class DeprecatedBundleWarning(DeprecationWarning):
    """
    The weak security bundle is being deprecated. Please bother your service
    provider to get them to stop using cross-signed roots.
    """


def where():
    f = os.path.dirname(__file__)

    return os.path.join(f, 'cacert.pem')


def old_where():
    warnings.warn(
        "The weak security bundle is being deprecated.",
        DeprecatedBundleWarning
    )
    f = os.path.dirname(__file__)
    return os.path.join(f, 'weak.pem')

if __name__ == '__main__':
    print(where())



# -*- coding: utf-8 -*-
"""
    wakatime.main
    ~~~~~~~~~~~~~

    Module entry point.

    :copyright: (c) 2013 Alan Hamlett.
    :license: BSD, see LICENSE for more details.
"""

from __future__ import print_function

import logging
import os
import sys
import traceback

pwd = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, os.path.dirname(pwd))
sys.path.insert(0, os.path.join(pwd, 'packages'))

from .__about__ import __version__
from .api import send_heartbeats
from .arguments import parse_arguments
from .compat import u, json
from .constants import SUCCESS, UNKNOWN_ERROR
from .logger import setup_logging

log = logging.getLogger('WakaTime')

from .heartbeat import Heartbeat
from .offlinequeue import Queue


def execute(argv=None):
    if argv:
        sys.argv = ['wakatime'] + argv

    args, configs = parse_arguments()

    setup_logging(args, __version__)

    try:
        heartbeats = []

        hb = Heartbeat(vars(args), args, configs)
        if hb:
            heartbeats.append(hb)
        else:
            log.debug(hb.skip)

        if args.extra_heartbeats:
            try:
                for extra_data in json.loads(sys.stdin.readline()):
                    hb = Heartbeat(extra_data, args, configs)
                    if hb:
                        heartbeats.append(hb)
                    else:
                        log.debug(hb.skip)
            except json.JSONDecodeError as ex:
                log.warning(u('Malformed extra heartbeats json: {msg}').format(
                    msg=u(ex),
                ))

        retval = send_heartbeats(heartbeats, args, configs)
        if retval == SUCCESS:
            queue = Queue(args, configs)
            offline_heartbeats = queue.pop_many()
            if len(offline_heartbeats) > 0:
                retval = send_heartbeats(offline_heartbeats, args, configs)

        return retval

    except:
        log.traceback(logging.ERROR)
        print(traceback.format_exc())
        return UNKNOWN_ERROR



# -*- coding: utf-8 -*-
"""
    wakatime.constants
    ~~~~~~~~~~~~~~~~~~

    Constant variable definitions.

    :copyright: (c) 2016 Alan Hamlett.
    :license: BSD, see LICENSE for more details.
"""

""" Success
Exit code used when a heartbeat was sent successfully.
"""
SUCCESS = 0

""" Api Error
Exit code used when the WakaTime API returned an error.
"""
API_ERROR = 102

""" Config File Parse Error
Exit code used when the ~/.wakatime.cfg config file could not be parsed.
"""
CONFIG_FILE_PARSE_ERROR = 103

""" Auth Error
Exit code used when our api key is invalid.
"""
AUTH_ERROR = 104

""" Unknown Error
Exit code used when there was an unhandled exception.
"""
UNKNOWN_ERROR = 105

""" Connection Error
Exit code used when there was proxy or other problem connecting to the WakaTime
API servers.
"""
CONNECTION_ERROR = 107

""" Max file size supporting line number count stats.
Files larger than this in bytes will not have a line count stat for performance.
Default is 2MB.
"""
MAX_FILE_SIZE_SUPPORTED = 2000000
